import asyncio
import logging
import socket
import threading

from decentnet.consensus.block_sizing import BLOCK_PREFIX_LENGTH_BYTES
from decentnet.consensus.blockchain_params import BlockchainParams
from decentnet.consensus.byte_conversion_constants import ENDIAN_TYPE
from decentnet.consensus.dev_constants import (BLOCK_SHARE_LOG_LEVEL,
                                               RUN_IN_DEBUG)
from decentnet.consensus.share_server_config import (INCOMING_CHANNEL_BYTE_LEN,
                                                     SUBSCRIBE_ACK,
                                                     UNSUB_BYTE_LEN,
                                                     UNSUB_BYTE_STRING)
from decentnet.modules.convert.byte_to_base64_utils import bytes_to_base64
from decentnet.modules.logger.log import setup_logger
from decentnet.modules.tcp.socket_functions import recv_all

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG if BLOCK_SHARE_LOG_LEVEL == logging.DEBUG else False, logger)
ADDR_ZERO = b"\x00" * 32


class BlockShareServer:
    def __init__(self, host='localhost', port1=8009, port2=8010):
        # First listener for storing addresses
        self.server_socket_1 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket_1.bind((host, port1))
        self.server_socket_1.listen(BlockchainParams.max_hosts)

        # Second listener for sending data to address
        self.server_socket_2 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket_2.bind((host, port2))
        self.server_socket_2.listen(BlockchainParams.max_hosts)

        # Dictionary to store address mappings
        self.address_dict = {}

        logger.debug(
            f"BlockShareServer started at {host}:{port1} for storing addresses and {host}:{port2} for sending data.")

    def listener_store_address(self):
        """ First listener: Store hash/socket pairs in dictionary """
        while True:
            conn, client_address = self.server_socket_1.accept()
            logger.debug(f"Accepting consumer {client_address}")
            threading.Thread(target=self.handle_consumer, args=(conn,), daemon=True,
                             name="Handling consumer").start()

    def handle_consumer(self, socket_obj: socket.socket):
        channel_bytes = socket_obj.recv(INCOMING_CHANNEL_BYTE_LEN)
        channel = bytes_to_base64(channel_bytes)

        if self.address_dict.get(channel) is None:
            logger.debug(f"Writing first subscriber {channel}")
        else:
            logger.debug(f"Overwriting {channel}")

        self.address_dict[channel] = socket_obj

        socket_obj.sendall(SUBSCRIBE_ACK)
        logger.debug(f"Current address dictionary looks like {tuple(self.address_dict.keys())}")
        try:
            cmd = socket_obj.recv(UNSUB_BYTE_LEN)
        except ConnectionResetError:
            logger.debug(f"Failed got {cmd}")
            cmd = UNSUB_BYTE_STRING

        if cmd == UNSUB_BYTE_STRING:
            logger.debug(f"Unsubscribed {channel}")
            self.address_dict.pop(socket_obj, None)
            socket_obj.close()

    def listener_send_data(self):
        """ Second listener: Send data to the socket identified by hash key """
        while True:
            conn, client_address = self.server_socket_2.accept()
            logger.debug(f"Accepting publisher {client_address}")
            threading.Thread(target=self.handle_publisher, args=(conn,), daemon=True,
                             name="Publisher handler").start()

    def handle_publisher(self, socket_obj: socket.socket):
        data_len = 1

        while data_len:
            try:
                raw_data, data_len = asyncio.run(recv_all(socket_obj))
                channel = bytes_to_base64(raw_data[:INCOMING_CHANNEL_BYTE_LEN])

                # Recompute length prefix to send only necessary data
                data = raw_data[INCOMING_CHANNEL_BYTE_LEN:]
                length_prefix = int.to_bytes(data_len - INCOMING_CHANNEL_BYTE_LEN,
                                             length=BLOCK_PREFIX_LENGTH_BYTES,
                                             byteorder=ENDIAN_TYPE, signed=False)

                logger.debug(f"Publisher sending data to {channel}")
                if (soc := self.address_dict.get(channel)) is not None:
                    soc.sendall(length_prefix + data)
            except ConnectionResetError:
                logger.debug("Publisher disconnected.")
                break

    def start(self):
        """ Start both listeners using threading """
        thread_1 = threading.Thread(target=self.listener_store_address, daemon=True, name="Address Store")
        # Start both listeners
        thread_1.start()

        # Wait for both threads to complete (they run indefinitely)

        self.listener_send_data()
        thread_1.join()


def start_block_share_server():
    server = BlockShareServer()
    server.start()


if __name__ == "__main__":
    start_block_share_server()



from enum import Enum


class NetworkCmd(Enum):
    """Enum of commands"""
    HANDSHAKE_ENCRYPTION = 0
    BROADCAST = 1  # Broadcast new connection
    SYNCHRONIZE = 2
    DISCONNECT_EDGE = 3  # R2R Comm extended
    REDELIVER_ON_CONNECT = 4



import secrets
import string

from decentnet.consensus.byte_conversion_constants import ENDIAN_TYPE


class SecurePasswordGenerator:
    def __init__(self, length=40, digits=True, upper=True, lower=True, special=True):
        self.length = length
        self.digits = digits
        self.upper = upper
        self.lower = lower
        self.special = special

    def generate(self):
        # Validate length
        if not isinstance(self.length, int) or self.length <= 0:
            raise ValueError("Password length must be a positive integer")

        char_sets = []

        # Collect possible character sets based on flags
        if self.digits:
            char_sets.append(string.digits)
        if self.upper:
            char_sets.append(string.ascii_uppercase)
        if self.lower:
            char_sets.append(string.ascii_lowercase)
        if self.special:
            char_sets.append(string.punctuation)

        if not char_sets:
            raise ValueError("At least one character type must be selected")

        # Flatten all characters into a single pool for random selection
        all_characters = ''.join(char_sets)

        # Ensure we can meet the requirement of including at least one char from each chosen set
        if self.length < len(char_sets):
            raise ValueError("Password length too short to include all chosen character types")

        # Introduce a random seed step: get a random seed from secrets
        # and use it to rotate the all_characters string
        seed_data = secrets.token_bytes(64)  # 512 bits of randomness
        seed_int = int.from_bytes(seed_data, ENDIAN_TYPE)
        # Rotate the character pool by a random offset
        offset = seed_int % len(all_characters)
        all_characters = all_characters[offset:] + all_characters[:offset]

        # Pick one guaranteed character from each selected set to ensure complexity
        password_chars = [secrets.choice(cset) for cset in char_sets]

        # Fill the rest with random characters from the entire pool
        remaining_length = self.length - len(password_chars)
        password_chars.extend(secrets.choice(all_characters) for _ in range(remaining_length))

        # Shuffle the resulting list of characters to ensure no pattern
        # SystemRandom is inherently cryptographically secure
        secrets.SystemRandom().shuffle(password_chars)

        # Join and return the final password
        password = ''.join(password_chars)
        return password



def e2b(value):
    """
    Convert a string or other types representing booleans to a Python boolean.

    Args:
        value: The value (usually from an environment variable) to convert to a boolean.

    Returns:
        bool: The converted boolean value.

    Raises:
        ValueError: If the value cannot be interpreted as a boolean.
    """
    truthy_values = {"true", "1", "t", "yes", "y", "on"}
    falsy_values = {"false", "0", "f", "no", "n", "off"}

    if isinstance(value, bool):
        return value

    if isinstance(value, str):
        value_lower = value.strip().lower()  # Strip any extra whitespace and normalize case
        if value_lower in truthy_values:
            return True
        elif value_lower in falsy_values:
            return False

    # Raise an error for anything that can"t be interpreted as a boolean
    raise ValueError(f"Cannot interpret {value} as a boolean.")



import logging

import cbor2

from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.consensus.routing_params import DEFAULT_CAPACITY
from decentnet.modules.forwarding.flow_net import FlowNetwork
from decentnet.modules.logger.log import setup_logger
from decentnet.modules.tasks_base.publisher import ClientBlockPublisher

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)


class ProcessingBlock:
    @staticmethod
    async def proces_broadcast_block(network: FlowNetwork, data: dict):
        logger.debug(
            f"Adding edge from broadcast data {data['pub']} => {data['target']}")
        await network.add_edge(data["pub"], data["target"], DEFAULT_CAPACITY)

    @staticmethod
    async def decrease_ttl_broadcast_block(data: dict):
        data["ttl"] -= 1
        return data

    @staticmethod
    async def process_disconnect_block(network: FlowNetwork, data: dict):
        xdata = cbor2.loads(data["data"])
        logger.debug(
            f"Removing edge from disconnect broadcast data {data['pub']} => {data['target']}")
        await network.rm_edge(xdata["s"], xdata["d"], DEFAULT_CAPACITY)
        await network.bulk_save_db()
        pipe1 = ClientBlockPublisher.pipes.pop(xdata["s"], None)
        ClientBlockPublisher.beams_ref.pop(xdata["s"], None)
        if pipe1:
            pipe1.close()



import asyncio
import hashlib
import logging
import os
import threading
from datetime import datetime, timezone
from socket import socket
from typing import Tuple

import cbor2
import nacl
from _socket import gaierror
from networkx.exception import NodeNotFound

from decentnet.consensus.beam2beam import HANDSHAKE_MSG
from decentnet.consensus.beam_constants import (BEAM_AES_ENCRYPTION_KEY_SIZE,
                                                BEAM_HASH_SIZE, USED_ALGORITHM)
from decentnet.consensus.blockchain_params import (
    SKIP_SIGNATURE_VERIFICATION_DEPTH, BlockchainParams)
from decentnet.consensus.byte_conversion_constants import ENDIAN_TYPE
from decentnet.consensus.cmd_enum import NetworkCmd
from decentnet.consensus.dev_constants import ENCRYPTION_DEBUG, RUN_IN_DEBUG, DEBUG_RELAY_USING
from decentnet.consensus.routing_params import DEFAULT_CAPACITY
from decentnet.modules.blockchain.block import Block
from decentnet.modules.blockchain.blockchain import Blockchain
from decentnet.modules.convert.byte_to_base64_utils import (base64_to_original,
                                                            bytes_to_base64)
from decentnet.modules.cryptography.asymmetric import AsymCrypt
from decentnet.modules.cryptography.symmetric_aes import AESCipher
from decentnet.modules.cryptography.symmetric_chacha import ChaChaCipher
from decentnet.modules.db.models import BeamTable
from decentnet.modules.forwarding.flow_net import FlowNetwork
from decentnet.modules.internal_processing.blocks import ProcessingBlock
from decentnet.modules.key_util.foreign_keys import KeyManagerForeign
from decentnet.modules.key_util.key_manager import KeyManager
from decentnet.modules.logger.log import setup_logger
from decentnet.modules.pow.difficulty import Difficulty
from decentnet.modules.tcp.client import TCPClient
from decentnet.modules.transfer.packager import Packager

logger = logging.getLogger(__name__)
setup_logger(RUN_IN_DEBUG, logger)


class Beam:
    conn_bc: Blockchain  # This is a blockchain for connection to relay
    comm_bc: Blockchain | None  # Blockchain for Beacon2Beacon communication

    def __init__(self, pub_key_id: int, pub_key_enc_id: int, target_key: str, do_metrics=False):
        """
        Beam for sending data transmissions between relay and beacon, always ip_port or client_socket is required
        :param pub_key_id:
        :param target_key:
        """
        self.do_metric = do_metrics
        self.hash = None
        self._lock = False
        self.comm_bc = None
        self.beam_id = None
        self.alive = False
        self.connected = False
        self.pub_key = None  # Public key under which this beam is identified to others
        self.pub_key_id = pub_key_id
        self.skip_verification = False
        _, o_pub_key = asyncio.run(KeyManager.retrieve_ssh_key_pair_from_db(
            self.pub_key_id))

        _, o_pub_enc_key = asyncio.run(KeyManager.retrieve_ssh_key_pair_from_db(pub_key_enc_id,
                                                                                can_encrypt=True))
        self.pub_enc_key = AsymCrypt.encryption_key_to_base64(o_pub_enc_key)
        self.pub_key = AsymCrypt.verifying_key_to_string(o_pub_key)
        self.target_key = target_key
        if DEBUG_RELAY_USING:
            if self.target_key != "NOT_KNOWN":
                logger.debug(
                    f"Using identity {self.pub_key} to receive data from {self.target_key} on thread {threading.current_thread().name} PID: {os.getpid()}")
            else:
                logger.debug(
                    f"Using identity {self.pub_key} to receive data from new beacons on thread {threading.current_thread().name} PID: {os.getpid()}")
        self.encryptor_relay: None | AESCipher = None
        self.encryptor_beacon: None | AESCipher = None
        self.client = None
        self.flow_net = FlowNetwork()
        self.encryption_init_complete = False

    def connect_using_address(self, ip_port: tuple):
        """
        Connects to address using ip_port tuple
        :param ip_port:
        :return:
        """
        if ip_port:
            try:
                self.client = TCPClient(ip_port[0], ip_port[1], ip_type=ip_port[2])
                self.hash = Beam.create_beam_hash(self.client.port)
                return True
            except gaierror:
                logger.debug(f"Failed to resolve {ip_port}")
                return False
            except ConnectionRefusedError:
                # logger.debug(f"No connection to {ip_port}")
                return False
        return False

    def connect_using_socket(self, client_socket: socket):
        """
        Connects to socket using client_socket
        :param client_socket:
        :return:
        """
        logger.debug("Reusing socket for beam")
        self.client = TCPClient(client_socket=client_socket)
        self.hash = Beam.create_beam_hash(self.client.port)
        return True

    async def initialize_incoming_transmission(self, genesis_block: Block) -> bool:
        """
        Connects incoming transmission and sends handshake block with encryption rules
        :param genesis_block:
        """
        self.conn_bc = Blockchain(pub_key_for_encryption=None,
                                  beam_id=self.beam_id, name="Connection blockchain",
                                  do_metric=self.do_metric)
        # await self.conn_bc.finish_blockchain_init()
        self.conn_bc.difficulty.compression_type = genesis_block.diff.compression_type
        logger.debug("Connection incoming with genesis %s" % genesis_block)
        await self.__connect_incoming(genesis_block, self.target_key)
        await self.init_comm_blockchain(True)
        self._save()
        return self.connected

    async def initialize_outgoing_transmission(self) -> bool:
        """
        Initializes outgoing transmission
        :return:
        """
        self.conn_bc = Blockchain(BlockchainParams.default_genesis_msg,
                                  pub_key_for_encryption=self.pub_enc_key,
                                  beam_id=self.beam_id, name="Connection blockchain",
                                  do_metric=False)
        await self.conn_bc.finish_blockchain_init()
        logger.debug(f"Connecting outgoing from {self.target_key}")
        await self.__connect_outbound_connection(self.pub_key_id, self.target_key)

        await self.init_comm_blockchain(True)

        self._save()
        return self.connected

    async def init_comm_blockchain(self, insert: bool, do_metric: bool = None):
        """
        Initializes communication blockchain from genesis block of connection blockchain
        """
        logger.debug(
            f"Initiation of communication blockchain with pub_enc {self.conn_bc.pub_key_for_encryption}")
        self.comm_bc = Blockchain(pub_key_for_encryption=self.conn_bc.pub_key_for_encryption,
                                  beam_id=self.beam_id, name="Communication Blockchain",
                                  do_metric=self.do_metric if do_metric is None else do_metric)
        await self.comm_bc.finish_blockchain_init()
        self.comm_bc.difficulty.compression_type = self.conn_bc.difficulty.compression_type
        if insert:
            if not await self.comm_bc.insert(self.conn_bc.chain[0]):
                raise Exception(
                    "Failed to initialize communication blockchain, genesis block may be invalid")

    def lock(self):
        self._lock = True

    def unlock(self):
        self._lock = False

    async def __connect_outbound_connection(self, pub_key_id, target_key):
        self.alive = len(self.conn_bc) > 0
        logger.debug(
            f"Sending Genesis block of Connection and Communication BC -> \n{self.conn_bc.get_last()}")
        response_handshake = await self.send_block(pub_key_id, target_key,
                                                   self.conn_bc.get_last())

        if not response_handshake:
            logger.warning(f"Got empty response, target {target_key}")
            self.close()
            raise Exception(f"Got empty response, target {target_key}")

        logger.debug("Response from Relay %s" % response_handshake)
        block = Block.from_bytes(response_handshake["data"])

        if not await self.conn_bc.insert(block):
            raise Exception("Invalid handshake block")

        block_data = cbor2.loads(block.data)

        if block_data["data"] != "CONNECTED":
            logger.warning("Connection failed, perhaps bad block")
            logger.debug("Block data %s" % block_data)
            self.close()
            raise Exception(f"Connection failed, perhaps bad block {block}")
        else:
            if self.target_key is None:
                self.target_key = response_handshake["pub"]
            logger.debug(f"Received new pubkey {self.target_key}")

            try:
                await self.process_handshake_block(block_data, True)
            except KeyError as e:
                logger.fatal("Wrong handshake block")
                logger.debug("Block data %s" % block_data)
                raise Exception(f"Wrong handshake block error {e}")

            await self.save_new_pub_key(self.pub_key, False, "New Beacon")
        self.connected = self.alive and self.encryptor_beacon

    async def process_handshake_block(self, block_data: dict, relay: bool, send_ack=True):
        """
        Process handshake block and save encryptor to either
        encryptor_relay if relay is True or encryptor_beacon if relay is False
        :param send_ack:
        :param block_data:
        :param relay:
        :return:
        """
        # Command does not have to be verified, block data are signed
        if not isinstance(block_data, dict):
            logger.error(f"Block data must be a dict not {block_data}")
            raise TypeError("Block data must be a dict")

        if block_data["cmd"] == NetworkCmd.HANDSHAKE_ENCRYPTION.value:
            logger.debug(
                f"Received handshake block | Using {block_data['algo']}-{block_data['bits']}")
            encryption_algorithm = block_data["algo"]
            encryption_bits = block_data["bits"]
            encrypted_key = base64_to_original(block_data["key"])

            private_key_from_db = await KeyManager.get_private_key(self.pub_enc_key)
            # TODO: from relay keys
            try:
                data_enc_pwd = AsymCrypt.decrypt_message(
                    private_key_from_db,
                    encrypted_key)
            except nacl.exceptions.CryptoError as e:
                logger.fatal(f"Failed to decrypt handshake data from related public key {self.pub_enc_key}")
                raise e

            await self.set_encryptors(data_enc_pwd, encryption_algorithm, encryption_bits, relay)

            logger.debug(f"Connected to {self.pub_key}")
            logger.debug("Decrypted key successfully")
            if relay:
                logger.info(
                    f"Relay connection beam from {self.pub_key} => {self.target_key} has {encryption_algorithm}-{encryption_bits} enabled!")
            else:
                logger.info(
                    f"Direct path beam from {self.pub_key} => {self.target_key} has {encryption_algorithm}-{encryption_bits} enabled!")
            if send_ack:
                await self.send_ack_connect(self.pub_key, self.target_key)

            return encryption_algorithm, encryption_bits
        return None

    async def set_encryptors(self, data_enc_pwd: bytes, encryption_algorithm: str, encryption_bits: int,
                             relay: bool):
        if encryption_algorithm == "AES":
            cipher = AESCipher(data_enc_pwd, encryption_bits)
            if relay:
                self.encryptor_relay = cipher
            else:
                self.encryptor_beacon = cipher
        elif encryption_algorithm == "CC20":
            cipher = ChaChaCipher(data_enc_pwd, encryption_bits)
            if relay:
                self.encryptor_relay = cipher
            else:
                self.encryptor_beacon = cipher
        else:
            raise NotImplementedError(
                f"Not implemented encryptor {encryption_algorithm}")

    async def __connect_incoming(self, genesis_block: Block, target_key: str):
        self.alive = await self.conn_bc.insert(genesis_block)

        if not self.alive:
            logger.warning("Failed to insert genesis block, closing connection.")
            self.close()

        block_data = cbor2.loads(genesis_block.data)
        logger.debug("Received genesis block data: %s" % block_data)
        received_encryption_pub_key = block_data["enc_pub_key"]
        new_diff = Difficulty.from_bytes(block_data["new_diff"].encode("ascii"))
        self.conn_bc.difficulty = new_diff

        # TODO: handshake block should not be sent if target key is known, as the connection is not only for connection
        logger.debug(
            f"Setting new public encryption key for Connection blockchain {received_encryption_pub_key}")
        self.conn_bc.pub_key_for_encryption = received_encryption_pub_key

        logger.debug("Sending handshake block")
        handshake_block, pwd = Blockchain.create_handshake_encryption_block_raw(
            received_encryption_pub_key, algorithm=USED_ALGORITHM, additional_data="CONNECTED")

        logger.debug("Waiting for ack...")

        response = await self.send_connection_data(handshake_block, ack=True)

        await self.set_encryptors(pwd, USED_ALGORITHM, BEAM_AES_ENCRYPTION_KEY_SIZE, relay=True)

        if not response:
            logger.error("Response is empty")
            self.close()
            raise Exception("Response is empty")
        if Packager.verify_dict_data(response["pub"], "status", response):
            logger.debug("Verified ack message, finalized connection.")
            if self.target_key is None:
                self.target_key = response["pub"]
        else:
            self.connected = False
            logger.error(f"Failed to verify ack message from {response['pub']}")
            logger.debug(f"Ack message obtained {response}")
            self.close()
            raise Exception("Failed to verify ack message, maybe MITM attack ?")

        await self.save_new_pub_key(received_encryption_pub_key,
                                    description=f"Key from {target_key}",
                                    can_encrypt=True)
        self.connected = self.alive and self.encryptor_beacon
        logger.info("Connected with ACK verification.")
        logger.debug(f"Incoming connection alive? {self.connected}")

    async def send_block(self, pub_key_id, target_key, block: Block, request_ack=True) -> dict:
        """Send block with needed pack structure

        Verification is skipped if blockchain gets long as we use AES encryption after 2nd block
        """
        data = await Packager.pack(pub_key_id,
                                   block, target_key, self.skip_verification)
        logger.debug("Sending block %s, will wait for ack..." % block)
        response = await self.client.send_message(data, request_ack)
        return response

    @classmethod
    def create_beam_hash(cls, port: int):
        """
        Create a beam hash using blake2b to identity beams
        :return:
        """
        timestamp = datetime.now(timezone.utc)
        timestamp_str = timestamp.isoformat()

        # Hash the timestamp string using BLAKE2b
        return bytes_to_base64(hashlib.blake2b(timestamp_str.encode("ascii"),
                                               salt=port.to_bytes(2,
                                                                  byteorder=ENDIAN_TYPE,
                                                                  signed=False),
                                               digest_size=BEAM_HASH_SIZE).digest())

    async def save_new_pub_key(self, pub_key: str, can_encrypt: bool, description: str):
        await KeyManagerForeign.save_to_db(pub_key, description, can_encrypt,
                                           self.client.host, self.client.port)

    async def send_ack_connect(self, source_key: str, target_key: str):
        pk, pub = await KeyManager.retrieve_ssh_key_pair_from_db(self.pub_key_id)
        pub_key = AsymCrypt.verifying_key_to_string(pub)
        found = True
        try:
            _, _ = self.flow_net.get_path(source_key, target_key)
        except NodeNotFound:
            found = False
        msg = {"status": "CONNECTED" if self.alive else "CLOSED",
               "node_found": found,
               "pub": pub_key,
               }

        signed_data = Packager.sign_dict_data(pk, "status", msg)

        logger.debug("Sending ack connect message")
        await self.client.send_message(cbor2.dumps(signed_data),
                                       False)
        return True

    async def send_connection_data(self, data_raw: bytes, ack: bool = False) -> dict:
        """
        Sending connection data,
        typically used for communication with relay or internal decentmesh communication
        :param data_raw:
        :param ack:
        :return:
        """
        return await self.__send_data(data_raw, ack, self.conn_bc, encryptor=self.encryptor_relay)

    async def send_communication_data(self, data_raw: bytes, ack: bool = False) -> dict:
        """
        Sending communication data, is used for data transmission for user
        This data is automatically encrypted if encryptor is set
        :param data_raw:
        :param ack:
        :return:
        """
        logger.debug("Trying to send communication data %s B" % (len(data_raw)))
        if len(self.comm_bc) == 0:
            logger.debug("Inserted genesis block from connection blockchain as communication is outbound")
            await self.comm_bc.insert(self.conn_bc.chain[0])

        if self.encryptor_beacon is None:
            logger.warning("! Communication data are sent unencrypted !")

        return await self.__send_data(data_raw, ack, self.comm_bc, encryptor=self.encryptor_beacon)

    async def __send_data(self, data_raw: str | bytes, ack: bool, blockchain: Blockchain,
                          encryptor: AESCipher | ChaChaCipher | None) -> dict:
        """
        Send Data with use of specified blockchain
        :param ack: This ack is used only for internal decent connect communication do not use for communication
        :param data_raw:
        :param encryptor: If encryptor is passed, data will be encrypted
        :return:
        """

        if encryptor:
            encrypted_data = encryptor.encrypt(data_raw)
            if RUN_IN_DEBUG:
                logger.debug(f"Encrypted data to {encrypted_data} B")
            if ENCRYPTION_DEBUG:
                logger.debug(f"PWD {encryptor.password} {encryptor.salt}")
                logger.debug(f"Sending encrypted data {encrypted_data} {data_raw}")
        else:
            encrypted_data = None

        block = blockchain.template_next_block(encrypted_data if encrypted_data else data_raw,
                                               blockchain.difficulty)

        block.mine()

        if not await blockchain.insert(block):
            raise Exception(f"Failed to insert block {block}")
        packaged_data = await Packager.pack(self.pub_key_id, block, self.target_key,
                                            self.skip_verification if self.comm_bc else False)

        response = await self.client.send_message(packaged_data, ack=ack)
        logger.debug(f"Response from sending message {response}")
        return response

    def decrypt_block_data(self, block: Block) -> dict:
        return cbor2.loads(self.encryptor_relay.decrypt(block.data))

    async def fetch_message(self, do_metrics=None, listen_for_target_broadcast_block=False) -> Tuple[
        Block, dict, bytes | bytearray | str]:
        """
            Waits for and processes a serialized block message from a connected client.

            This method listens for an incoming message,
            deserializes it into a Block instance, and inserts it into an incoming blockchain.
            If no client connection exists, it raises an exception.
            It relies on Packager.unpack to process the specific message format and extract block data,
            while logging the received block for debugging.

               Returns:
                   Block: The Block instance deserialized from the received message.
                   dict: Unpacked original data in dict
                   bytes: Raw data

               Raises:
                   Exception: If no client connection is established prior to invocation.

               Side Effects:
                   - Reads a message from the network through the `client.receive_message` interface.
                   - Inserts the received block into the `blockchain`

               Note:
                   This method does validate the received block as it is inserting block.
        """
        if self._lock:
            raise Exception("You can not receive data from beam that is used by Relay!")

        if not self.client:
            raise Exception("No connection is established.")

        # Ignore broadcast blocks of other that got connected
        incoming_raw = await self.client.receive_message(decode=False)
        verified, incoming_data, verified_csig = Packager.unpack(incoming_raw, self.skip_verification)
        Packager.check_verified(incoming_data, verified)
        block = Block.from_bytes(incoming_data["data"])

        # WORKAROUND TODO: FIX
        if block.index == 0 and len(self.comm_bc) == 1:
            # Workaround if comm blockchain is initialized with improper genesis
            await self.init_comm_blockchain(False, do_metric=do_metrics)
            await self.comm_bc.insert(block)

            if not self.skip_verification:
                self.skip_verification = self.check_if_verification_skippable()

        cbor_loaded_data = cbor2.loads(block.data)
        if cbor_loaded_data == 0:
            if RUN_IN_DEBUG:
                logger.debug("Block data is not directly loadable, checking for commands")

            if self.encryptor_beacon and RUN_IN_DEBUG:
                logger.debug(f"Received encrypted block #{block.index}, passing data without loading...")
            elif incoming_data["cmd"] == NetworkCmd.REDELIVER_ON_CONNECT.value:
                real_block = Block.from_bytes(block.data)
                if listen_for_target_broadcast_block:
                    return real_block, incoming_data, incoming_raw
                else:
                    logger.debug(
                        "Received Redeliver block from Relay but silently skipping due to ignore flag...")
                    return await self.fetch_message(do_metrics=do_metrics)
            else:
                raise Exception(f"Failed to load block data {block}")

        if verified_csig is not None:  # Broadcast is faulty or handled incorrectly
            # Verifying command signature
            Packager.check_verified(incoming_data, verified_csig)
            logger.debug(
                f"Handling connection data {incoming_data['cpub']} {NetworkCmd(incoming_data['cmd']).name}")
            await self.handle_connection_data(incoming_data, block)
            return await self.fetch_message(do_metrics=do_metrics)
        elif block.index == 0:
            logger.debug(f"Received enc pub key genesis block from {incoming_data['pub']}")
            await self.save_new_pub_key(cbor_loaded_data["enc_pub_key"],
                                        description=f"Key from {incoming_data['pub']}",
                                        can_encrypt=True)
            if listen_for_target_broadcast_block and incoming_data["pub"] == self.target_key:
                logger.debug(
                    f"Received broadcast block from target {self.target_key}, target is ready to receive data!")
                return block, incoming_data, incoming_raw
            else:
                return await self.fetch_message(do_metrics=do_metrics)

        if incoming_data["pub"] == self.target_key and block.index < len(self.comm_bc):
            logger.debug(f"Obtained genesis block twice from target {self.target_key}")
            logger.debug(f"Current Genesis {self.comm_bc.chain[0]}")
            logger.debug(f"Received genesis {block}")

            return await self.fetch_message(do_metrics=do_metrics)

        if self.encryptor_beacon is None:
            logger.debug("Expecting handshake block, trying to process")
            await self.process_handshake_block(cbor_loaded_data, False, False)
            self.comm_bc.difficulty = block.diff
            await self.comm_bc.insert(
                block)  # Genesis block differs, need to get genesis block from sender maybe broadcast ?
            if not self.skip_verification:
                self.skip_verification = self.check_if_verification_skippable()
            await self.send_communication_data(cbor2.dumps(HANDSHAKE_MSG))
            logger.debug("Waiting for incoming traffic")
            return await self.fetch_message(do_metrics=do_metrics)

        logger.debug(f"Received block {block}")

        # TODO: add requesting difficulty to avoid spam
        self.comm_bc.difficulty = block.diff
        await self.comm_bc.insert(block)

        if not self.skip_verification:
            self.skip_verification = self.check_if_verification_skippable()

        return block, incoming_data, incoming_raw

    def check_if_verification_skippable(self):
        return len(self.comm_bc) > SKIP_SIGNATURE_VERIFICATION_DEPTH

    async def handle_connection_data(self, incoming_data: dict, incoming_block: Block):
        logger.debug(f"> Processing received connection block {incoming_data}")
        self.conn_bc.difficulty = incoming_block.diff
        await self.conn_bc.insert(incoming_block)

        if incoming_block.index == 0 and len(self.conn_bc) > 0:
            if (cmd := incoming_data.get("cmd", None)) and cmd == NetworkCmd.BROADCAST.value:
                await self.flow_net.add_edge(incoming_data["pub"], self.target_key,
                                             DEFAULT_CAPACITY)
                logger.debug("Adding edge from broadcast block")
            else:
                logger.debug(
                    f"Received connection block {incoming_block}, processing not implemented")
        if incoming_data["cmd"] == NetworkCmd.DISCONNECT_EDGE.value:
            await ProcessingBlock.process_disconnect_block(self.flow_net, incoming_data)

    def set_communication_difficulty(self, difficulty: Difficulty):
        self.comm_bc.difficulty = difficulty

    async def init_secure_b2b_connection(self, target_enc_key: str):
        """
        Init secure beacon to beacon connection over communication stream
        :param target_enc_key: Target encryption public key
        """
        block, data_enc_pwd = self.comm_bc.create_handshake_encryption_block_dict(
            target_enc_key)
        handshake_block_bytes = self.comm_bc.convert_handshake_block_dict_to_bytes(block)
        await self.send_communication_data(handshake_block_bytes, False)
        encryption_algorithm = block["algo"]
        encryption_bits = block["bits"]
        await self.set_encryptors(data_enc_pwd, encryption_algorithm, encryption_bits, False)

    def close(self):
        self.client.close()

    def _save(self):
        if self.comm_bc.save or self.conn_bc.save:
            asyncio.create_task(
                BeamTable.save(self, self.pub_key, self.target_key, self.conn_bc.id, self.comm_bc.id),
                name="SaveBeamTable"
            )



from typing import List

from sqlalchemy import select

from decentnet.modules.db.base import session_scope
from decentnet.modules.db.constants import USING_ASYNC_DB
from decentnet.modules.db.models import AliveBeam


async def get_alive_beams() -> List[AliveBeam]:
    if USING_ASYNC_DB:
        async with session_scope() as session:
            # Perform an asynchronous query to retrieve all living beams
            result = await session.execute(select(AliveBeam).where(AliveBeam.ready))
            all_alive_beams = result.scalars().all()
        return all_alive_beams
    else:
        with session_scope() as session:
            # Perform a synchronous query to retrieve all living beams
            result = session.execute(select(AliveBeam).where(AliveBeam.ready))
            all_alive_beams = result.scalars().all()
        return all_alive_beams



import numpy as np


class IcosaBalancer:
    IN_LATENCY_TARGET = 50.0
    OUT_LATENCY_TARGET = 50.0
    BYTE_SIZE_TARGET = 1500.0
    TOLERANCE_LATENCY = 0.05  # 5% tolerance for latencies
    TOLERANCE_BYTE_SIZE = 0.1  # 10% tolerance for byte size
    MAX_ITERATIONS = 50

    @staticmethod
    def balance_client_latencies(clients):
        data_matrix = np.array([[client['in_latency'], client['out_latency'], client['byte_size']]
                                for client in clients.values()])

        targets = np.array([IcosaBalancer.IN_LATENCY_TARGET, IcosaBalancer.OUT_LATENCY_TARGET,
                            IcosaBalancer.BYTE_SIZE_TARGET])
        tolerances = np.array([IcosaBalancer.TOLERANCE_LATENCY * IcosaBalancer.IN_LATENCY_TARGET,
                               IcosaBalancer.TOLERANCE_LATENCY * IcosaBalancer.OUT_LATENCY_TARGET,
                               IcosaBalancer.TOLERANCE_BYTE_SIZE * IcosaBalancer.BYTE_SIZE_TARGET])

        adjustments_matrix = np.zeros_like(data_matrix)

        for iteration in range(IcosaBalancer.MAX_ITERATIONS):
            adjustments_made = False

            # Calculate current values with adjustments
            final_values = data_matrix + adjustments_matrix
            deviations = final_values - targets
            abs_deviations = np.abs(deviations)

            # Check if each metric is within tolerance for each client
            within_tolerance = abs_deviations <= tolerances

            # If all values are within tolerance, stop adjusting
            if within_tolerance.all():
                break

            # Adjust extreme values
            for i in range(data_matrix.shape[0]):
                for j in range(3):  # For in_latency, out_latency, byte_size
                    if not within_tolerance[i, j]:  # Adjust only if outside tolerance
                        adjustment_step = -0.1 * deviations[i, j]  # Incremental step towards target
                        adjustments_matrix[i, j] += adjustment_step
                        adjustments_made = True

            # Stop if no adjustments were made in an iteration
            if not adjustments_made:
                break

        # Map final adjustments to each client
        client_ids = list(clients.keys())
        adjustments = {client_ids[i]: {
            'in_latency_adjustment': adjustments_matrix[i, 0],
            'out_latency_adjustment': adjustments_matrix[i, 1],
            'byte_size_adjustment': adjustments_matrix[i, 2]
        } for i in range(len(client_ids))}

        return adjustments



import io
import logging
import os
import socket
import threading
from typing import Optional

from decentnet.consensus.block_sizing import BLOCK_PREFIX_LENGTH_BYTES
from decentnet.consensus.byte_conversion_constants import ENDIAN_TYPE
from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.consensus.tcp_params import RECV_BUFFER_SIZE, SEND_BUFFER_SIZE
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)


def set_sock_properties(socket_obj: socket.socket):
    socket_obj.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, RECV_BUFFER_SIZE)
    socket_obj.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, SEND_BUFFER_SIZE)
    socket_obj.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)


async def recv_prefix(sock: socket.socket, length: int) -> memoryview:
    """
    Receive exactly `length` bytes from `sock`.
    Returns a memoryview over a bytearray that holds the data.
    Raises ConnectionError if the socket closes before we get all bytes.
    """
    if length <= 0:
        return memoryview(bytearray(0))  # or just return memoryview(b'')

    buf = bytearray(length)  # Pre-allocate exactly `length` bytes
    view = memoryview(buf)
    received = 0

    while received < length:
        n = sock.recv_into(view[received:], length - received)
        if n == 0:
            raise ConnectionError("Socket closed before reading required bytes.")
        received += n

    return view


async def recv_all(socket_obj: socket.socket, host: Optional[str] = None,
                   port: Optional[int] = None, length_prefix_size=BLOCK_PREFIX_LENGTH_BYTES) -> (
        tuple)[bytes, int]:
    """Receive all bytes of a message up to total_bytes from a socket."""
    if RUN_IN_DEBUG:
        logger.debug(
            f"Thread {threading.current_thread().name} in PID {os.getpid()} is reading socket {socket_obj}")

    try:
        length_prefix = await recv_prefix(socket_obj, length_prefix_size)
    except (ConnectionError, ConnectionResetError, ConnectionAbortedError,
            ConnectionRefusedError) as e:
        logger.debug(f"Beacon {host}:{port} forcefully disconnected")
        raise e

    total_bytes = int.from_bytes(length_prefix, ENDIAN_TYPE, signed=False)
    if total_bytes:
        if RUN_IN_DEBUG:
            logger.debug(
                f"Incoming message will have length {total_bytes} B |"
                f" prefix {length_prefix.hex()} | Thread: {threading.current_thread().name}")
    else:
        return b'', 0

    buffer = io.BytesIO()
    data_len = 0

    while data_len < total_bytes:
        # Calculate the remaining bytes to receive
        remaining_bytes = total_bytes - data_len

        # Receive up to the remaining number of bytes
        frame = socket_obj.recv(remaining_bytes)

        if not frame:
            # No more data is being sent; possibly the connection was closed
            logger.debug(
                f"Data is not being sent, closing connection... Remaining {remaining_bytes} Bytes "
                f"Thread: {threading.current_thread().name} | Buffer contained {buffer.getvalue()}")
            break

        # Write the received frame into the BytesIO buffer
        written_bytes = buffer.write(frame)
        data_len += written_bytes

    # Retrieve the accumulated data
    data = buffer.getvalue()
    # Close the buffer to free up memory
    buffer.close()
    return data, data_len



T_COST_BYTE_SIZE = 4
M_COST_BYTE_SIZE = 4
P_COST_BYTE_SIZE = 1
N_BITS_BYTE_SIZE = 1
HASH_LEN_CHARS_BYTE_SIZE = 1
COMPRESSION_LEVEL_BYTE_SIZE = 1
COMPRESSION_TYPE_BYTE_SIZE = 1
EXPRESS_POW_BYTE_SIZE = 1

# Chunk slicing to calculate only once per import
M_COST_CHUNK_SLICE = T_COST_BYTE_SIZE + M_COST_BYTE_SIZE
P_COST_CHUNK_SLICE = M_COST_CHUNK_SLICE + P_COST_BYTE_SIZE
N_BITS_CHUNK_SLICE = P_COST_CHUNK_SLICE + N_BITS_BYTE_SIZE
HASH_LEN_CHARS_CHUNK_SLICE = N_BITS_CHUNK_SLICE + HASH_LEN_CHARS_BYTE_SIZE
COMPRESSION_LEVEL_CHUNK_SLICE = HASH_LEN_CHARS_CHUNK_SLICE + COMPRESSION_LEVEL_BYTE_SIZE
COMPRESSION_TYPE_CHUNK_SLICE = COMPRESSION_LEVEL_CHUNK_SLICE + COMPRESSION_TYPE_BYTE_SIZE
EXPRESS_POW_CHUNK_SLICE = COMPRESSION_TYPE_CHUNK_SLICE + EXPRESS_POW_BYTE_SIZE



import asyncio
import logging
import os
from contextlib import asynccontextmanager, contextmanager
from functools import lru_cache
from pathlib import Path

from sqlalchemy import create_engine as sync_create_engine, text
from sqlalchemy.orm import sessionmaker

from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.consensus.local_config import (DATABASE_URL_ASYNC,
                                              DATABASE_URL_SYNC, DB_FILENAME)
from decentnet.modules.db import constants
from decentnet.modules.db.models import Base
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)

try:
    import greenlet
    from sqlalchemy.ext.asyncio import async_sessionmaker, create_async_engine

    constants.USING_ASYNC_DB = True
except ImportError:
    constants.USING_ASYNC_DB = False
    logger.warning("Greenlet not installed using synchronous sqlite instead")


@lru_cache()
def get_root_dir() -> Path:
    return Path(os.path.abspath(__file__)).parent.parent.parent.parent


async def init_db(eng):
    if constants.USING_ASYNC_DB:
        # Async initialization
        async with eng.connect() as conn:
            await conn.run_sync(Base.metadata.create_all)
            await conn.execute(text("PRAGMA journal_mode=WAL"))
            await conn.execute(text("PRAGMA busy_timeout = 60"))
    else:
        # Sync initialization as fallback
        with eng.connect() as conn:
            Base.metadata.create_all(conn)
            await conn.execute(text("PRAGMA journal_mode=WAL"))
            await conn.execute(text("PRAGMA busy_timeout = 60"))

        # Switch between async or sync engine depending on Greenlet availability


db_file = get_root_dir() / DB_FILENAME
if constants.USING_ASYNC_DB:
    engine = create_async_engine(DATABASE_URL_ASYNC, echo=False)
else:
    engine = sync_create_engine(DATABASE_URL_SYNC, echo=False)

asyncio.run(init_db(engine))

if constants.USING_ASYNC_DB:
    # Modify session_scope to handle both async and sync sessions
    @asynccontextmanager
    async def session_scope():
        """Provide a transactional scope around a series of operations asynchronously."""
        async_session = async_sessionmaker(
            bind=engine,
            expire_on_commit=False
        )()
        try:
            yield async_session
            await async_session.commit()
        except Exception as e:
            await async_session.rollback()
            raise e
        finally:
            await async_session.close()

else:
    @contextmanager
    def session_scope():
        """Provide a transactional scope around a series of operations synchronously."""
        sync_session = sessionmaker(
            bind=engine,
            expire_on_commit=False
        )()
        try:
            yield sync_session
            sync_session.commit()
        except Exception as e:
            sync_session.rollback()
            raise e
        finally:
            sync_session.close()



"""
Debug Constants
"""
import logging
import os

from decentnet.consensus.env_convert_funcs import e2b

DEBUG_DATA_SERIALIZATION = e2b(os.getenv("DEBUG_DATA_SERIALIZATION", False))
DEBUG_SEED_CONNECTIONS = e2b(os.getenv("DEBUG_SEED_CONNECTIONS", False))
DEBUG_RELAY_USING = e2b(os.getenv("DEBUG_RELAY_USING", False))
RUN_IN_DEBUG = e2b(os.getenv("DEBUG", False))
LOG_LEVEL = os.getenv("DEBUG_LEVEL", logging.DEBUG)
ENCRYPTION_DEBUG = e2b(os.getenv("ENCRYPTION_DEBUG", False))
METRICS = e2b(os.getenv("METRICS", False))

DEBUG_TIMING = e2b(os.getenv("DEBUG_TIMING", False))
DEBUG_LZ4 = e2b(os.getenv("DEBUG_LZ4", False))
BLOCK_SHARE_LOG_LEVEL = os.getenv("BLOCK_SHARE_LOG_LEVEL", logging.ERROR)
SEEDS_AGENT_LOG_LEVEL = os.getenv("SEEDS_AGENT_LOG_LEVEL", logging.INFO)
COMPRESSION_LOG_LEVEL = os.getenv("COMPRESSION_LOG_LEVEL", logging.WARNING)
R2R_LOG_LEVEL = os.getenv("R2R_LOG_LEVEL", logging.INFO)
BLOCK_ERROR_DATA_LOG_LEN = 50



import logging
import os
import socket

from decentnet.consensus.block_sizing import BLOCK_PREFIX_LENGTH_BYTES
from decentnet.consensus.byte_conversion_constants import ENDIAN_TYPE
from decentnet.consensus.dev_constants import (BLOCK_SHARE_LOG_LEVEL,
                                               RUN_IN_DEBUG)
from decentnet.modules.convert.byte_to_base64_utils import base64_to_original
from decentnet.modules.logger.log import setup_logger
from decentnet.modules.tcp.socket_functions import set_sock_properties_common

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG if BLOCK_SHARE_LOG_LEVEL == logging.DEBUG else False, logger)


class Publisher:
    """
    Publisher class that sends data to a specific channel (hash key) via the server.
    """

    def __init__(self):
        self.server_host = "127.0.0.1"
        self.server_port = 8010  # 8010 ports are used for sending data to consumers
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        set_sock_properties_common(self.socket)
        self.socket.connect((self.server_host, self.server_port))

    async def __send_data(self, data):
        data_len = len(data)
        length_prefix = int.to_bytes(data_len, length=BLOCK_PREFIX_LENGTH_BYTES,
                                     byteorder=ENDIAN_TYPE, signed=False)
        logger.debug(
            "Outgoing message prefix %s | %s bytes, with prefix %s B" % (
                length_prefix.hex(), data_len, BLOCK_PREFIX_LENGTH_BYTES + data_len))

        self.socket.sendall(length_prefix + data)

    async def publish_message(self, channel: str, data: bytes) -> None:
        """
        Send a message to a specific channel (hash key) via the server.

        Args:
            channel (str): The hash key of the subscriber to which the message is to be sent.
            data (str): The message to send to the subscriber.
        """
        logger.debug("Publishing message to channel %s PID: %s}" % (channel, os.getpid()))
        channel_bytes = base64_to_original(channel)
        await self.__send_data(channel_bytes + data)



import asyncio
import logging
import multiprocessing
import random
from time import sleep
from typing import Optional

import netifaces

from decentnet.consensus.dev_constants import (RUN_IN_DEBUG,
                                               SEEDS_AGENT_LOG_LEVEL, DEBUG_SEED_CONNECTIONS)
from decentnet.consensus.net_constants import SEED_NODES_IPV4, SEED_NODES_IPV6
from decentnet.modules.comm.beacon import Beacon
from decentnet.modules.cryptography.asymmetric import AsymCrypt
from decentnet.modules.forwarding.flow_net import FlowNetwork
from decentnet.modules.key_util.key_manager import KeyManager
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)
logger.setLevel(SEEDS_AGENT_LOG_LEVEL)

setup_logger(RUN_IN_DEBUG, logger)


class SeedsAgent:

    def __init__(self, host: str, current_port: int, do_metric=False):
        self.beams = []
        self.network = FlowNetwork()
        self.current_port = current_port
        self.host = host
        self.pub_key_enc_id = 4
        self.host_ips = self._get_ip_addresses()
        self.do_metric = do_metric
        multiprocessing.Process(target=self.connect_to_seeds,
                                name="Seed connection agent", daemon=True).start()

    @classmethod
    def _get_ip_addresses(cls):
        ip_addresses = []
        interfaces = netifaces.interfaces()

        for interface in interfaces:
            addrs = netifaces.ifaddresses(interface)
            # Check for IPv4 addresses
            if netifaces.AF_INET in addrs:
                for addr in addrs[netifaces.AF_INET]:
                    ip_addresses.append(addr['addr'])

        return ip_addresses

    def connect_to_seeds(self, seeds_array: Optional[list] = None):
        if seeds_array:
            seed_nodes_copy = seeds_array
        else:
            seed_nodes_copy = SEED_NODES_IPV4 + SEED_NODES_IPV6

        for node_v4 in list(seed_nodes_copy):
            if (node_v4[0] == self.host and node_v4[1] == self.current_port) or (
                    node_v4[0] in self.host_ips and node_v4[1] == self.current_port):
                # Skip current host to avoid loopback
                continue
            try:
                beacon = Beacon(node_v4[0], node_v4[1], 1, node_v4[2], do_metric=self.do_metric)
                beam = beacon.create_beam("NOT_KNOWN", self.pub_key_enc_id, loopback=True)
                if beam is None or beam.client is None:
                    continue

                self.beams.append(beam)
                _, o_pub_key = asyncio.run(KeyManager.retrieve_ssh_key_pair_from_db(
                    beam.pub_key_id))
                pub_key = AsymCrypt.verifying_key_to_string(o_pub_key)
                # Invalid target key
                asyncio.run(self.network.add_edge(pub_key,
                                                  beam.target_key, 10))
                # Thread(target=beam.wait, name=f"SEED_BEAM_{node_v4}",
                #       daemon=True).start()
                logger.info(f"[OK] Seed {node_v4[0]}:{node_v4[1]} connected!")
                seed_nodes_copy.remove(node_v4)
                # TODO: This should be at every beam
                # TODO: ask for genesis block so seed can use this beacon as a relay

            except (ConnectionError, ConnectionResetError, ConnectionRefusedError,
                    ConnectionAbortedError, TimeoutError, OSError):
                if DEBUG_SEED_CONNECTIONS:
                    logger.debug(f"{node_v4[0]}:{node_v4[1]} failed to connect")
            finally:
                sleep(random.randint(1, 5))

        if len(seed_nodes_copy):
            if DEBUG_SEED_CONNECTIONS:
                logger.debug(
                    f"Will run seed connecting for {len(seed_nodes_copy)} seeds again")
            sleep(10)
            self.connect_to_seeds(seed_nodes_copy)



import asyncio
import logging
import time
from datetime import datetime, UTC
from multiprocessing import Pipe
from typing import Any, Callable, Coroutine

import cbor2
import networkx as nx
from networkx import NetworkXNoPath, NodeNotFound
from sqlalchemy import select, delete
from sqlalchemy.exc import IntegrityError

from decentnet.consensus.blockchain_params import \
    SKIP_SIGNATURE_VERIFICATION_DEPTH, BlockchainParams
from decentnet.consensus.cmd_enum import NetworkCmd
from decentnet.consensus.dev_constants import METRICS, RUN_IN_DEBUG
from decentnet.consensus.relay_config import (RELAY_DEFAULT_ENCRYPTION_KEY_ID,
                                              RELAY_DEFAULT_SIGNING_KEY_ID,
                                              RELAY_FREQUENCY_DB_ALIVE_UPDATE)
from decentnet.consensus.routing_params import (DEFAULT_CAPACITY,
                                                MAX_ROUTE_LENGTH)
from .block_assembly import sign_block, assemble_disconnect_data
from ..blockchain.block import Block
from ..comm.beam import Beam
from ..comm.db_funcs import get_alive_beams
from ..cryptography.asymmetric import AsymCrypt
from ..db.base import session_scope
from ..db.constants import USING_ASYNC_DB
from ..db.models import AliveBeam, NodeInfoTable, Mail
from ..forwarding.flow_net import FlowNetwork
from ..internal_processing.blocks import ProcessingBlock
from ..key_util.key_manager import KeyManager
from ..logger.log import setup_logger
from ...consensus.r2rcomm import BLOCK_COMMAND_PREFIX

if METRICS:
    from ..monitoring.metric_server import ping, send_metric

from ..serializer.serializer import Serializer
from ..tasks_base.publisher import ClientBlockPublisher
from ..tcp.socket_functions import recv_all
from ..timer.relay_counter import RelayCounter
from ..timer.timer import Timer
from ..transfer.packager import Packager

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)

AsyncCallbackType = Callable[[str, str], Coroutine[Any, Any, Any]]


class Relay:

    def __init__(self, client_socket, beam_pipe_comm: dict, beam_msg_queue: dict,
                 callback: AsyncCallbackType | None = None,
                 beam: Beam | None = None):
        """
        Creates a passive relay that will listen on socket and relay incoming beams
        :param client_socket:
        :param beam_pipe_comm: Dict for beam synchronization
        :param beam_msg_queue: Dict for beam message queue
        :param callback:
        """
        ClientBlockPublisher(beam_msg_queue)
        self.skip_verification = False
        global RELAY_PROCESS_INSTANCE
        RELAY_PROCESS_INSTANCE = self

        self.do_metrics = asyncio.run(ping()) if METRICS else False
        if self.do_metrics:
            logger.debug("Metrics will be collected.. OK")
        else:
            logger.debug("Metrics in Relay will not be collected.. FAIL")

        self.relay_pub_key_bytes = None
        self.alive = True
        self.beam_pipe_comm = beam_pipe_comm
        self.received_beam_pub_key = None
        self.socket = client_socket
        self.client_ip = client_socket.getpeername()
        self.__callback = callback
        self.public_key_id = RELAY_DEFAULT_SIGNING_KEY_ID
        self.public_key_enc_id = RELAY_DEFAULT_ENCRYPTION_KEY_ID

        logger.debug(f"Initial Connection {self.client_ip}")
        self.local_port = self.client_ip[1]

        logger.info("Waiting for genesis block from new sender...")

        try:
            request = asyncio.run(recv_all(self.socket, self.client_ip[0], self.local_port))[0]
        except (ConnectionError, ConnectionResetError, ConnectionAbortedError,
                ConnectionRefusedError, OSError):
            request = None

        if not request:
            raise ConnectionError

        self.network = FlowNetwork()

        verified, unpacked_request, verified_csig = Packager.unpack(request)
        Packager.check_verified(unpacked_request, verified)

        result = asyncio.run(self.execute_network_cmd_n_init_relay_key_bytes(unpacked_request, verified_csig))
        if result:
            return

        block = Block.from_bytes(unpacked_request["data"])  # Difficulty is checked when adding to blockchain

        self.received_beam_pub_key = pub_key = unpacked_request[
            "pub"]  # Key, which is received from unknown entity
        self.target_key = target_key = unpacked_request["target"]  # Key of mine or next destination

        self.init_pipes(beam_msg_queue, pub_key, target_key)

        if not beam:
            # Create a pipe for a relay beam communication between processes
            beam_pipe_comm[pub_key] = Pipe()

            self.beam = Beam(self.public_key_id, self.public_key_enc_id, target_key, self.do_metrics)
            # This is for each process to be able to get its relay address which is Relay public ckey
            global RELAY_PUB_KEY
            RELAY_PUB_KEY = pub_key

            self.beam.connect_using_socket(client_socket)
            asyncio.run(self.beam.initialize_incoming_transmission(block))
        else:
            self.beam = beam

        self.beam.lock()

        asyncio.run(self.complete_init_steps(block, pub_key, request, target_key, unpacked_request))

    async def execute_network_cmd_n_init_relay_key_bytes(self, unpacked_request, verified_csig):
        if verified_csig is not None:
            await self.execute_network_cmd(unpacked_request, verified_csig)
            return True  # Indicate that the function exited early

        _, self.relay_pub_key_bytes = await KeyManager.retrieve_ssh_key_pair_from_db(self.public_key_id)
        return False  # Indicate that the function continued normally

    async def send_all_undelivered_mail(self, target_key: str):
        async def __prepare_block_for_publish():
            next_block = self.beam.comm_bc.template_next_block(mail.block,
                                                               BlockchainParams.low_diff_argon)
            packed_data = await Packager.pack(self.public_key_id, next_block, target_key)
            verified, data, _ = Packager.unpack(packed_data)
            _data = await Packager.add_cmd(data, owner_key_id=self.public_key_id,
                                           cmd=NetworkCmd.REDELIVER_ON_CONNECT.value)
            serialized = Serializer.serialize_data(_data["pub"],
                                                   _data["sig"],
                                                   _data["data"],
                                                   _data["target"],
                                                   _data["cmd"],
                                                   _data["csig"],
                                                   _data["cpub"])
            return serialized

        send_pipe = self.beam_pipe_comm.get(target_key)
        if not send_pipe:
            return

        if USING_ASYNC_DB:
            async with session_scope() as session:
                await self._delete_expired_mail()
                result = await session.execute(
                    select(Mail).where(Mail.target == target_key)
                )
                mails = result.scalars().all()
                logger.debug(f"Will send {len(mails)} undelivered blocks from mailbox to {target_key}")

                for mail in mails:
                    await session.delete(mail)
                    serialized_broadcast = await __prepare_block_for_publish()
                    await ClientBlockPublisher.publish_message(target_key, serialized_broadcast)
        else:
            with session_scope() as session:
                await self._delete_expired_mail()
                result = session.execute(
                    select(Mail).where(Mail.target == target_key)
                )
                mails = result.scalars().all()
                logger.debug(f"Will send {len(mails)} undelivered blocks from mailbox to {target_key}")

                for mail in mails:
                    session.delete(mail)
                    serialized_broadcast = await __prepare_block_for_publish()
                    await ClientBlockPublisher.publish_message(target_key, serialized_broadcast)

    async def complete_init_steps(self, block: Block, pub_key: str, request: bytes, target_key: str,
                                  unpacked_request: dict):
        _, o_pub_key = await KeyManager.retrieve_ssh_key_pair_from_db(self.public_key_id)
        signing_pub_key = AsymCrypt.verifying_key_to_string(o_pub_key)
        await self.network.add_edge(signing_pub_key,
                                    pub_key, DEFAULT_CAPACITY)

        # TODO: look for undelivered mail

        if target_key != "NOT_KNOWN":
            await self.record_alive_beam(target_key, True)
            await self.send_all_undelivered_mail(pub_key)
            await self.relay_message_by_one(block, unpacked_request, request)

        # await self.network.add_edge(target_key, pub_key, DEFAULT_CAPACITY)
        if block.index == 0:
            logger.info(f"Adding connected Beacon {pub_key}")
            await self.save_beacon(pub_key)

            asyncio.create_task(
                self.broadcast_connected(block, pub_key, unpacked_request),
                name=f"Broadcasting of {pub_key}"
            )
        else:
            logger.info(f"Updating beacon connection {pub_key}")
            await self.update_beacon_connection(pub_key)
            await self.network.add_edge(pub_key,
                                        KeyManager.key_to_base64(target_key), None)
        if not self.beam.alive:
            self.beam.close()
            logger.warning("INVALID BLOCK, Closed connection")
        else:
            await self.beam.save_new_pub_key(pub_key, False, "New Beacon")

    @classmethod
    async def record_alive_beam(cls, pub_key: str, ready: bool):
        if USING_ASYNC_DB:
            async with session_scope() as session:
                alive_beam = (await session.execute(
                    select(AliveBeam).where(AliveBeam.pub_key == pub_key))).scalar_one_or_none()
                if alive_beam:
                    alive_beam.ready = ready
                else:
                    ab = AliveBeam(pub_key=pub_key, ready=ready)
                    session.add(ab)
        else:
            with session_scope() as session:
                alive_beam = (await session.execute(
                    select(AliveBeam).where(AliveBeam.pub_key == pub_key))).scalar_one_or_none()
                if alive_beam:
                    alive_beam.ready = ready
                else:
                    ab = AliveBeam(pub_key=pub_key, ready=ready)
                    session.add(ab)

    @classmethod
    def init_pipes(cls, beam_msg_queue, pub_key, target_key):
        if beam_msg_queue.get(pub_key, None) is None:
            beam_msg_queue[pub_key] = Pipe()
        if beam_msg_queue.get(target_key, None) is None:
            beam_msg_queue[target_key] = Pipe()

    @classmethod
    async def _add_block_to_mailbox(cls, serialized_block: bytes, request: bytes | None = None, target=None):
        if USING_ASYNC_DB:
            async with session_scope() as session:
                await cls._delete_expired_mail()
                session.add(Mail(block=serialized_block, request=request, target=target))
        else:
            with session_scope() as session:
                await cls._delete_expired_mail()
                session.add(Mail(block=serialized_block, request=request, target=target))

    @classmethod
    async def _delete_expired_mail(cls):
        now = datetime.now(UTC)
        if USING_ASYNC_DB:
            async with session_scope() as session:
                await session.execute(delete(Mail).where(Mail.expire_at < now))
        else:
            with session_scope() as session:
                session.execute(delete(Mail).where(Mail.expire_at < now))

    async def broadcast_connected(self, genesis_block: Block, connected_pub_key: str,
                                  unpacked_genesis: dict):
        """

        :param genesis_block:
        :param connected_pub_key: Pub key of the connected beacon needs to be base64
        :param unpacked_genesis: Unpacked genesis block in dict format
        """
        if genesis_block.index == 0:
            all_alive_beams = await get_alive_beams()

            _data = await Packager.add_cmd(
                unpacked_genesis, self.public_key_id, NetworkCmd.BROADCAST.value
            )
            ttl = _data.get("ttl", 0)
            logger.debug(
                f"Broadcasting connected beacon {connected_pub_key} to connected beams TTL {ttl}")
            block_with_signature = {
                "data": _data["data"],
                "sig": _data["sig"],
                "pub": _data["pub"]
            }
            # Template next connection block with broadcast block
            bb_data = cbor2.dumps(block_with_signature)
            broadcast_block = self.beam.conn_bc.template_next_block(bb_data, self.beam.conn_bc.difficulty)
            # broadcast_block.mine()
            if not await self.beam.conn_bc.insert(broadcast_block):
                raise Exception("Failed to insert broadcast block")
            broadcast_block_signature, broadcast_block_bytes = await sign_block(self.public_key_id,
                                                                                broadcast_block)
            serialized_broadcast = Serializer.serialize_data(
                self.relay_pub_key_bytes,
                broadcast_block_signature,
                broadcast_block_bytes,
                _data["target"],
                _data["cmd"],
                _data["csig"],
                _data["cpub"],
                ttl
            )

            await self._add_block_to_mailbox(broadcast_block_bytes, serialized_broadcast)

            # Broadcasting connection to other beams

            if len(all_alive_beams):
                await self._broadcast_data(all_alive_beams, serialized_broadcast)
            else:
                logger.info(f"No one to broadcast to {connected_pub_key}")

    @classmethod
    async def _broadcast_data(cls, all_alive_beams: list, serialized_broadcast: bytes):
        for beam in all_alive_beams:
            # TODO: Skip broadcast to current
            logger.info(f"  broadcasting connection to {beam.pub_key}")
            await ClientBlockPublisher.publish_message(beam.pub_key, serialized_broadcast)

    async def update_beacon_connection(self, pub_key):
        await Relay.update_beacon(pub_key)
        await self.do_callback(pub_key, "ping")

    async def do_callback(self, pub_key: str, action: str):
        if self.__callback:
            await self.__callback(pub_key, action)

    async def do_relaying(self, t: Timer, relay_counter: RelayCounter):
        """
        This function provides a single relay of request from Relay loop
        :param t: Timer
        :param relay_counter: Counter of relays
        :return: Request size (if 0 connection closed)
        """
        self.alive = True
        logger.debug(
            f"Waiting for data from {self.received_beam_pub_key} for relaying on {self.client_ip}...")
        try:
            request, request_size = (await recv_all(self.socket, self.client_ip[0], self.local_port))
        except (ConnectionError, ConnectionResetError, ConnectionAbortedError,
                ConnectionRefusedError, OSError):
            return await self.erase_all_edge_connections()

        if not request:
            return await self.erase_all_edge_connections()

        if self.do_metrics:
            await send_metric("prom_data_received", request_size)
        logger.info(f"Relay Connection {self.socket.getpeername()}")
        try:
            await self.relay_request(request, t, relay_counter)
        except (cbor2.CBORDecodeError, cbor2.CBORDecodeValueError):
            logger.error(f"Unable to decode: {request}")
            logger.error("Suggesting disconnect")
            self.beam.unlock()
            return await self.erase_all_edge_connections()

        return request_size

    async def erase_all_edge_connections(self):
        all_alive_beams = await get_alive_beams()

        for beam in all_alive_beams:
            if beam.pub_key == self.beam.pub_key:
                logger.debug(f"Skipped broadcasting disconnect to {beam.pub_key}")
                continue
            serialized_broadcast = cbor2.dumps(await assemble_disconnect_data(self.public_key_id,
                                                                              self.beam.pub_key,
                                                                              self.received_beam_pub_key))
            if beam.pub_key in self.beam_pipe_comm.keys():
                # Send disconnect details to another process with the correct blockchain for block assembly
                self.beam_pipe_comm[beam.pub_key][1].send(BLOCK_COMMAND_PREFIX + serialized_broadcast)
            else:
                logger.warning(f"Missing {beam.pub_key} in pipe comm dict unable sending disconnect")

        self.alive = False
        await self.network.rm_edge(self.beam.pub_key, self.received_beam_pub_key)
        return 0

    async def _execute_network_cmd(self, data: dict, cmd_value: int):
        await self.do_callback(data["pub"], NetworkCmd(cmd_value).name)
        if cmd_value == NetworkCmd.BROADCAST.value:
            await ProcessingBlock.proces_broadcast_block(self.network, data)
            return await ProcessingBlock.decrease_ttl_broadcast_block(data)
        elif cmd_value == NetworkCmd.SYNCHRONIZE.value:
            block = Block.from_bytes(data["data"])
            await self.beam.conn_bc.insert(block)

            if not self.skip_verification:
                self.skip_verification = await self.check_if_verification_needed()
        elif cmd_value == NetworkCmd.DISCONNECT_EDGE.value:
            await ProcessingBlock.process_disconnect_block(self.network, data)
            return await ProcessingBlock.decrease_ttl_broadcast_block(data)
        return None

    async def check_if_verification_needed(self):
        res = len(self.beam.comm_bc) > SKIP_SIGNATURE_VERIFICATION_DEPTH
        if res:
            logger.debug(
                f"Beam {self.beam.pub_key} reached signature verification depth of "
                f"{SKIP_SIGNATURE_VERIFICATION_DEPTH}.. Skipping verification")
        return res

    async def execute_network_cmd(self, unpacked_request: dict, verified_csig: bool):
        Packager.check_verified(unpacked_request, verified_csig)
        cmd_value = unpacked_request["cmd"]

        if RUN_IN_DEBUG:
            logger.debug(f"Received verified cmd {NetworkCmd(cmd_value)}")

        changed_data = await self._execute_network_cmd(unpacked_request, cmd_value)

        if changed_data:
            await self.rebroadcast(changed_data)

    async def rebroadcast(self, changed_data):
        if RUN_IN_DEBUG:
            logger.debug(f"Rebroadcasting data TTL {changed_data.get("ttl")}")
        all_alive_beams = await get_alive_beams()
        serialized_broadcast = cbor2.dumps(changed_data)
        await self._broadcast_data(all_alive_beams, serialized_broadcast)

    async def relay_request(self, request: bytes, t: Timer, relay_counter: RelayCounter):
        """
        Relay specified request
        :param request: Request to relay
        :param t: Timer
        :param relay_counter: RelayCounter
        :return:
        """
        t.stop()

        verified, data, verified_csig = Packager.unpack(request, self.skip_verification)

        if not self.skip_verification:
            Packager.check_verified(data, verified)
            self.skip_verification = await self.check_if_verification_needed()

        if verified_csig is not None:
            await self.execute_network_cmd(data, verified_csig)
            return

        block = Block.from_bytes(data["data"])

        block.signature = data["sig"]
        beacon_pub_key = data["pub"]

        self.beam.comm_bc.difficulty = block.diff
        insert_res = await self.beam.comm_bc.insert(block)

        if relay_counter.count % RELAY_FREQUENCY_DB_ALIVE_UPDATE == 0:
            await self.update_beacon_connection(beacon_pub_key)
            relay_counter.reset()

        if not insert_res:
            logger.error(f"Failed to insert block, closing connection... {self.client_ip}")
            self.socket.close()
            self.alive = False

        if block.index > 0:
            try:
                await self.relay_message_by_one(block, data, request)
            except nx.NetworkXNoPath:
                logger.warning(f"No path between {self.received_beam_pub_key} => {data['target']}")

        block_process_time_timer = t.stop()
        if RUN_IN_DEBUG:
            logger.debug(f"Total block process time: {block_process_time_timer} ms")

        if self.do_metrics:
            await send_metric("prom_block_process_time", block_process_time_timer)

        relay_counter.count += 1

    async def relay_message_by_one(self, block: Block, data: dict, request: bytes):
        path = None
        send_pipe = None
        store_to_mailbox = False
        process_pub_key = None
        try:
            path, capacity = self.network.get_path(self.received_beam_pub_key, data["target"])
            path_len = len(path)
            if self.do_metrics:
                await send_metric("prom_block_path_len", path_len)

            if path_len == 1:
                logger.debug("Path too short")
                return
            if path and path_len > MAX_ROUTE_LENGTH:
                logger.info("Maximum path exceeded, connecting to closer relay for better latency...")
                # TODO: connect to the closer relay to make shorter path
            logger.debug(f"Found path {path} and capacity {capacity} for block {block.index}")
            if send_pipe := self.beam_pipe_comm.get(path[-1], False):
                process_pub_key = path[-1]
            elif send_pipe := self.beam_pipe_comm.get(path[1], False):
                process_pub_key = path[1]
            else:
                store_to_mailbox = True
                if len(path):
                    process_pub_key = path[-1]
        except NodeNotFound:
            logger.warning(f"Node {data['target']} not found")
            store_to_mailbox = True
        except NetworkXNoPath:
            logger.warning(f"Path to {data['target']} not found")
            store_to_mailbox = True

        if store_to_mailbox:
            await self._add_block_to_mailbox(data["data"], request, process_pub_key)
            logger.debug(f"Added block to mailbox for {process_pub_key} and quit")
            return

        if RUN_IN_DEBUG:
            logger.debug(
                f"Publishing from {self.received_beam_pub_key} message to {process_pub_key} on path {path}")

        if not send_pipe:
            logger.warning(f"No pipe to send message to {process_pub_key}")
            return

        # Syncing inter relay blockchain
        send_pipe[1].send(data["data"])

        await ClientBlockPublisher.publish_message(process_pub_key, request)

    @classmethod
    async def update_beacon(cls, pub_key: str):
        if USING_ASYNC_DB:
            async with session_scope() as session:
                # Perform an asynchronous query to get the beacon
                result = await session.execute(
                    select(NodeInfoTable).where(NodeInfoTable.pub_key == pub_key)
                )
                beacon = result.scalar_one_or_none()

                if beacon:
                    beacon.last_ping = time.time()
        else:
            with session_scope() as session:
                # Perform a synchronous query to get the beacon
                result = session.execute(
                    select(NodeInfoTable).where(NodeInfoTable.pub_key == pub_key)
                )
                beacon = result.scalar_one_or_none()

                if beacon:
                    beacon.last_ping = time.time()

    async def disconnect_beacon(self, port: int, pub_key: str):
        # Call the disconnect callback asynchronously
        if USING_ASYNC_DB:
            # Call the disconnect callback asynchronously
            await self.do_callback(pub_key, "disconnect")

            async with session_scope() as session:
                # Perform an asynchronous query to find the record to delete
                result = await session.execute(
                    select(NodeInfoTable).where(
                        (NodeInfoTable.port == port) & (NodeInfoTable.pub_key == pub_key)
                    )
                )
                record_to_delete = result.scalar_one_or_none()

                if record_to_delete:
                    await session.delete(record_to_delete)
        else:
            # Call the disconnect callback synchronously
            await self.do_callback(pub_key, "disconnect")

            with session_scope() as session:
                # Perform a synchronous query to find the record to delete
                result = session.execute(
                    select(NodeInfoTable).where(
                        (NodeInfoTable.port == port) & (NodeInfoTable.pub_key == pub_key)
                    )
                )
                record_to_delete = result.scalar_one_or_none()

                if record_to_delete:
                    session.delete(record_to_delete)

    async def save_beacon(self, pub_key: str):
        # Define a helper function to perform the database operations
        def add_beacon_to_session():
            client_ip = self.client_ip
            bdb = NodeInfoTable(
                ipv4=client_ip[0],
                port=client_ip[1],
                pub_key=pub_key,
            )
            session.add(bdb)

        # Define a helper function to handle commit and update
        async def commit_or_update_beacon():
            try:
                # Commit the transaction
                if USING_ASYNC_DB:
                    await session.commit()
                else:
                    session.commit()
            except IntegrityError:
                logger.debug(f"Updating connected node {self.client_ip[0]}:{self.client_ip[1]}")
                if USING_ASYNC_DB:
                    await session.rollback()
                    await Relay.update_beacon(pub_key)
                else:
                    session.rollback()
                    await Relay.update_beacon(pub_key)

        # Perform the callback
        if USING_ASYNC_DB:
            await self.do_callback(pub_key, "connect")

            async with session_scope() as session:
                add_beacon_to_session()
                await commit_or_update_beacon()
        else:
            await self.do_callback(pub_key, "connect")

            with session_scope() as session:
                add_beacon_to_session()
                await commit_or_update_beacon()


RELAY_PUB_KEY: str | None = None
RELAY_PROCESS_INSTANCE: Relay | None = None



import logging
from typing import Tuple

import cbor2

from decentnet.consensus.cmd_enum import NetworkCmd
from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.modules.blockchain.block import Block
from decentnet.modules.cryptography.asymmetric import AsymCrypt
from decentnet.modules.logger.log import setup_logger
from decentnet.modules.serializer.serializer import Serializer
from decentnet.modules.transfer.packager import Packager

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)


async def assemble_disconnect_data(public_key_id: int, source: str,
                                   dest: str):
    """
    Assemble serialized block
    :param public_key_id:
    :param source:
    :param dest:
    :return:
    """
    __data = {"s": source, "d": dest}
    # disconnect_block = blockchain.template_next_block(cbor2.dumps(__data))
    # disconnect_block.mine()

    _data = await Packager.add_cmd(
        __data, public_key_id, NetworkCmd.DISCONNECT_EDGE.value
    )
    _data["cpub"] = AsymCrypt.verifying_key_to_string(_data["cpub"])
    # ttl = _data.get("ttl", 0)
    # logger.debug(
    #    f"Broadcasting disconnect {source} => {dest} to connected beams TTL {ttl}")
    # if not await blockchain.insert(disconnect_block):
    #    raise Exception("Failed to insert disconnect block")
    #
    # block_signature, block_bytes = await sign_block(public_key_id, disconnect_block)

    # serialized_block = Serializer.serialize_data(
    #    relay_pub_key_bytes,
    #    block_signature,
    #    block_bytes,
    #    "ALL",
    #    _data["cmd"],
    #    _data["csig"],
    #    _data["cpub"],
    #    ttl
    # )
    return _data


async def assemble_disconnect_block(public_key_id, relay_pub_key_bytes, blockchain, r2r_data) -> (
        Tuple)[bytes, Block]:
    """
    Assemble serialized block
    :param r2r_data:
    :param blockchain:
    :param relay_pub_key_bytes:
    :param public_key_id:
    :return:
    """
    disconnect_block = blockchain.template_next_block(cbor2.dumps(r2r_data))
    disconnect_block.mine()

    ttl = r2r_data.get("ttl", 0)
    if not await blockchain.insert(disconnect_block):
        raise Exception("Failed to insert disconnect block")
    #
    block_signature, block_bytes = await sign_block(public_key_id, disconnect_block)

    serialized_block = Serializer.serialize_data(
        relay_pub_key_bytes,
        block_signature,
        block_bytes,
        "ALL",
        r2r_data["cmd"],
        r2r_data["csig"],
        AsymCrypt.public_key_from_base64(r2r_data["cpub"], False),
        ttl
    )
    return serialized_block, disconnect_block


async def sign_block(public_key_id: int, block: Block) -> Tuple[bytes, bytes]:
    """Will sign block with current public key id of Relay"""
    block_bytes = await block.to_bytes()
    block_signature = await Packager.sign_block(
        public_key_id,
        block_bytes
    )
    return block_signature, block_bytes



import random
import time
import unittest
from multiprocessing import Manager, Pipe, Process

from decentnet.consensus.block_sizing import MAXIMUM_BLOCK_SIZE


def generate_messages(num_messages: int) -> list:
    """
    Generate random messages beforehand, providing verbose output for each message.

    Args:
        num_messages (int): The number of random messages to generate.

    Returns:
        list: A list of random byte messages.
    """
    messages = []
    for i in range(num_messages):
        size = random.randint(1, MAXIMUM_BLOCK_SIZE)
        message = random.randbytes(size)
        messages.append(message)

        # Verbose output for each message
        print(f"Generated message {i + 1}/{num_messages} - Size: {size} bytes")

    print(f"Total {num_messages} messages generated.")
    return messages


def consumer_process(pipe, num_messages):
    """
    Function for consumer process to consume messages.

    Args:
        pipe: The read end of the pipe.
        num_messages: Number of messages to consume.
    """
    consumer_conn = pipe[0]  # The consumer reads from pipe[0]
    for _ in range(num_messages):
        consumer_conn.recv()


def producer_process(pipe, messages):
    """
    Function for producer process to publish pre-generated messages.

    Args:
        pipe: The write end of the pipe.
        messages: List of pre-generated messages.
    """
    producer_conn = pipe[1]  # The producer writes to pipe[1]
    for message in messages:
        producer_conn.send(message)  # Send the message through the pipe


class PerformanceTest(unittest.TestCase):

    def test_performance(self):
        """
        Test the performance of the Producer/Consumer message passing using pipes.
        Measures both the latency (time it takes for a message to reach the consumer)
        and the throughput (messages per second).
        """

        # Number of messages to test throughput
        num_messages = 1000
        print("Generating messages...")
        manager = Manager()
        messages = manager.list(generate_messages(num_messages))

        # Create a Pipe for communication
        pipe = Pipe()

        # Start the timer for latency and throughput measurement
        start_time = time.perf_counter_ns()

        # Create consumer and producer processes
        consumer_proc = Process(target=consumer_process, args=(pipe, num_messages))
        producer_proc = Process(target=producer_process, args=(pipe, messages))

        # Start both processes
        consumer_proc.start()
        producer_proc.start()

        # Wait for both processes to finish
        producer_proc.join()
        consumer_proc.join()

        # End time after processes complete
        end_time = time.perf_counter_ns()

        # Calculate the time taken and throughput
        total_time = end_time - start_time

        # Latency per message (in nanoseconds)
        latency_per_message = total_time / num_messages

        # Throughput calculation: number of messages per second
        throughput = num_messages / (total_time / 1_000_000_000)

        # Output in milliseconds (convert from nanoseconds)
        print(f"Total time for {num_messages} messages: {total_time / 1_000_000:.6f} ms")
        print(f"Average latency per message: {latency_per_message / 1_000_000:.6f} ms")
        print(f"Throughput: {throughput:.6f} messages per second")


if __name__ == '__main__':
    unittest.main()



from concurrent.futures.thread import ThreadPoolExecutor
from typing import Optional

from Crypto.Cipher import ChaCha20_Poly1305
from Crypto.Random import get_random_bytes
from argon2.low_level import Type, hash_secret

from decentnet.consensus.beam_constants import BEAM_AES_ENCRYPTION_KEY_SIZE


class ChaChaCipher:
    # Class-level constants
    SALT_SIZE = 12  # 12-byte salt
    NONCE_SIZE = 12  # Standard 12-byte nonce
    TAG_SIZE = 16  # Authentication tag size
    DERIVE_KEY_ONCE = True  # Derive the key only once (toggleable)
    CHUNK_SIZE = 1024 * 1024  # 1 MB for a parallelization threshold

    def __init__(self, password: bytes, key_size: int = BEAM_AES_ENCRYPTION_KEY_SIZE,
                 salt: Optional[bytes] = None):
        """
        Initializes the ChaChaCipher with a secure password and optional salt.
        """
        self.password = password
        self.salt = salt if salt is not None else get_random_bytes(self.SALT_SIZE)
        self.key_size = key_size

        # Derive the key only once if DERIVE_KEY_ONCE is True
        if self.DERIVE_KEY_ONCE:
            self.key = self.derive_key(self.password, self.salt)

    def derive_key(self, password: bytes, salt: bytes) -> bytes:
        """
        Derives a cryptographic key using Argon2id with exactly 32 bytes (256 bits).
        """
        derived_key = hash_secret(
            password,
            salt,
            time_cost=1,
            memory_cost=8,
            parallelism=1,
            hash_len=self.key_size,
            type=Type.ID
        )
        return derived_key[:32]

    def encrypt_chunk(self, chunk: bytes, nonce: bytes) -> bytes:
        """
        Encrypts a single chunk using ChaCha20-Poly1305.
        """
        cipher = ChaCha20_Poly1305.new(key=self.key, nonce=nonce)
        ciphertext, tag = cipher.encrypt_and_digest(chunk)
        return nonce + tag + ciphertext

    def decrypt_chunk(self, chunk: bytes, key: bytes) -> bytes:
        """
        Decrypts a single chunk using ChaCha20-Poly1305.
        """
        nonce = chunk[:self.NONCE_SIZE]
        tag = chunk[self.NONCE_SIZE:self.NONCE_SIZE + self.TAG_SIZE]
        encrypted_data = chunk[self.NONCE_SIZE + self.TAG_SIZE:]
        cipher = ChaCha20_Poly1305.new(key=key, nonce=nonce)
        return cipher.decrypt_and_verify(encrypted_data, tag)

    def encrypt(self, plaintext: bytes) -> bytes:
        """
        Encrypts the plaintext using ChaCha20-Poly1305.
        The output includes salt, nonce, tag, and ciphertext.
        Parallelizes the process if data is larger than CHUNK_SIZE.
        """
        if len(plaintext) <= self.CHUNK_SIZE:
            nonce = get_random_bytes(self.NONCE_SIZE)
            return self.salt + self.encrypt_chunk(plaintext, nonce)

        # Split data into chunks and process in parallel
        chunks = [plaintext[i:i + self.CHUNK_SIZE] for i in range(0, len(plaintext), self.CHUNK_SIZE)]
        with ThreadPoolExecutor() as executor:
            nonces = [get_random_bytes(self.NONCE_SIZE) for _ in chunks]
            encrypted_chunks = list(executor.map(self.encrypt_chunk, chunks, nonces))
        return self.salt + b"".join(encrypted_chunks)

    def decrypt(self, ciphertext: bytes) -> bytes:
        """
        Decrypts the ciphertext using ChaCha20-Poly1305.
        Parallelizes the process if data is larger than CHUNK_SIZE.
        """
        salt = ciphertext[:self.SALT_SIZE]
        encrypted_data = ciphertext[self.SALT_SIZE:]

        # Derive key if necessary
        key = self.key if self.DERIVE_KEY_ONCE else self.derive_key(self.password, salt)

        # Check if data is small enough for single decryption
        if len(encrypted_data) <= self.CHUNK_SIZE + self.NONCE_SIZE + self.TAG_SIZE:
            return self.decrypt_chunk(encrypted_data, key)

        # Split encrypted data into chunks and process in parallel
        chunk_size = self.CHUNK_SIZE + self.NONCE_SIZE + self.TAG_SIZE
        chunks = [encrypted_data[i:i + chunk_size] for i in range(0, len(encrypted_data), chunk_size)]
        with ThreadPoolExecutor() as executor:
            decrypted_chunks = list(executor.map(self.decrypt_chunk, chunks, [key] * len(chunks)))
        return b"".join(decrypted_chunks)



import asyncio
import logging

import cbor2
import click
import rich
from rich.panel import Panel

from decentnet.consensus.blockchain_params import BlockchainParams
from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.interface.basic import BasicInterface
from decentnet.modules.comm.beacon import Beacon
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)


@click.group()
def client():
    pass


@client.command()
@click.argument("owner_key_id")
@click.argument("host", type=str)
@click.argument("port", type=int)
@click.argument("target-key", type=str, required=False,
                default="Ah70wYetFbhu9X/nrFQsFXTW/3kE4GKeRfhQ+1os1rcL")  # "Target public signing key"
@click.argument("target-enc-key", type=str, required=False, default=None)  # "Target encryption key"
@click.option("--wait", "-w", is_flag=True, default=False)
def start(owner_key_id, host: str, port: int, target_key: str, target_enc_key: str, wait: bool) -> None:
    beacon = Beacon(host, port, owner_key_id)
    client_str = f"Identity {beacon.pub_key}\nConnecting to:\nHost: {host}\nPort: {port}\nTargeting: {target_key}"
    rich.print(Panel(client_str, title="Client Info"))
    beam = beacon.create_beam(target_key, 4, False)
    if beam is None:
        logger.error("No connection could be made, is relay running ?")
        return

    # Todo: it looks like that client gets genesis block and then it gets broadcast block which results in broadcast block not able to insert
    if not wait:
        # Get target enc pub key

        # This is used for the second layer of security between client and client as a last encryption layer
        # Client is expecting the first block to be handshake block
        # enc_target_key = session.query(ForeignKeys).where(
        #    ForeignKeys.description == f"Key from {target_key}").first()
        if target_enc_key is None:
            logger.error("No public key for encryption")
            exit(1)
        logger.debug("> Communication genesis block is \n%s" % (beam.comm_bc.get_last()))
        asyncio.run(beam.fetch_message(listen_for_target_broadcast_block=True))
        # Set lower difficulty for handshake block
        beam.comm_bc.difficulty = BlockchainParams.low_diff_argon
        block, pwd = beam.comm_bc.create_handshake_encryption_block_dict(
            target_enc_key)
        # Send handshake block for direct communication without relay being able to decrypt
        handshake_block_bytes = beam.comm_bc.convert_handshake_block_dict_to_bytes(block)
        asyncio.run(beam.send_communication_data(handshake_block_bytes, False))
        asyncio.run(beam.process_handshake_block(block, False, False))
        logger.debug("Waiting for ack handshake")
        data = asyncio.run(beam.fetch_message())[0].data
        logger.debug("MSG: %s" % beam.encryptor_beacon.decrypt(bytes(data)))
        asyncio.run(beam.send_communication_data("Hello :) incom".encode("ascii")))
        asyncio.run(beam.send_communication_data("Hello :) incom2".encode("ascii")))

    if wait:
        logger.info("Waiting for incoming beams")
        for _ in range(2):
            incoming_block = asyncio.run(beam.fetch_message())[0]
            loaded = beam.encryptor_beacon.decrypt(bytes(incoming_block.data))
            logger.debug(f"Received MSG {loaded}")
        beam.close()


@client.command()
@click.argument("owner_key_id")
@click.argument("owner_key_enc_id")
@click.argument("target-key", type=str, required=False,
                default="Ah70wYetFbhu9X/nrFQsFXTW/3kE4GKeRfhQ+1os1rcL")  # "Target public signing key"
@click.argument("target-enc-key", type=str, required=False, default=None)  # "Target encryption key"
def testsend(owner_key_id, owner_key_enc_id, target_key: str, target_enc_key: str) -> None:
    bi = BasicInterface(target_key, target_enc_key)
    bi.init_sending(owner_key_id, owner_key_enc_id)
    print(bi.encrypted_send)
    bi.send(cbor2.dumps({"data": "Hello"}))


@client.command()
@click.argument("owner_key_id")
@click.argument("owner_key_enc_id")
@click.argument("target-key", type=str, required=False,
                default="Ah70wYetFbhu9X/nrFQsFXTW/3kE4GKeRfhQ+1os1rcL")  # "Target public signing key"
@click.argument("target-enc-key", type=str, required=True)  # "Target encryption key"
def testrec(owner_key_id, owner_key_enc_id, target_key: str, target_enc_key: str) -> None:
    bi = BasicInterface(target_key, target_enc_key)
    bi.init_receiving(owner_key_id, owner_key_enc_id)

    print(asyncio.run(bi.fetch()))



import asyncio
import logging
from collections import deque
from time import sleep

from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.consensus.metrics_constants import (
    MAX_REQUEST_QUEUE_LEN, REQUEST_MAX_FREQUENCY,
    WAIT_FOR_REQUEST_IN_EMPTY_QUEUE_DELAY)
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)


class ReqQueue:
    _requeue: deque = None
    item_count: int = 0

    @classmethod
    def append(cls, func):
        if cls._requeue is not None:
            if cls.item_count >= MAX_REQUEST_QUEUE_LEN:
                logger.debug("Metric request was lost, queue too full")
                return
            cls._requeue.append(func)
            cls.item_count += 1

    @classmethod
    def init_queue(cls):
        cls._requeue = deque(maxlen=MAX_REQUEST_QUEUE_LEN)

    @classmethod
    def do_requests(cls, init=True):
        logger.debug("Starting request queue")
        if init:
            cls.init_queue()
        while True:
            try:
                asyncio.run(cls._requeue.pop())
            except IndexError:
                sleep(WAIT_FOR_REQUEST_IN_EMPTY_QUEUE_DELAY)
            else:
                cls.item_count -= 1
                sleep(REQUEST_MAX_FREQUENCY)



import time
from datetime import datetime, timedelta, UTC
from sqlite3 import IntegrityError

import sqlalchemy
from sqlalchemy import (REAL, Boolean, Column, ForeignKey, Integer,
                        LargeBinary, String, UniqueConstraint, select, BLOB, DateTime)
from sqlalchemy.orm import declarative_base, mapped_column

from decentnet.modules.db.constants import USING_ASYNC_DB

Base = declarative_base()


class BlockTable(Base):
    __tablename__ = 'block'

    block_id = Column(Integer, primary_key=True, autoincrement=True, index=True)
    blockchain_id = Column(Integer,
                           ForeignKey("blockchain.blockchain_id", ondelete="CASCADE"),
                           index=True)
    index = Column(Integer, nullable=False, index=True)
    hash = Column(LargeBinary, nullable=False)
    previous_hash = Column(LargeBinary, nullable=False)
    timestamp = Column(REAL, nullable=False, default=time.time())
    data = Column(LargeBinary, nullable=False)
    nonce = Column(Integer, nullable=False)

    def __repr__(self):
        return f"<Block(id={self.block_id}, hash='{self.hash}', previous_hash='{self.previous_hash}', " \
               f"timestamp='{self.timestamp}', data='{self.data}')>"


class BlockSignatureTable(Base):
    __tablename__ = "block_signature"
    block_id = Column(Integer, ForeignKey("block.block_id", ondelete="CASCADE"),
                      primary_key=True, index=True)
    signature = Column(String(256), nullable=False)


class BlockchainTable(Base):
    __tablename__ = "blockchain"
    beam_id = Column(Integer, ForeignKey("beam.id", ondelete="CASCADE"), index=True)
    blockchain_id = Column(Integer, primary_key=True, autoincrement=True)
    timestamp = Column(REAL, default=time.time(), nullable=False)
    version = Column(String(16), nullable=False)

    difficulty = Column(String(128), nullable=False)


class ForeignKeys(Base):
    __tablename__ = 'foreign_key'
    id = Column(Integer, primary_key=True, autoincrement=True)
    public_key = Column(String(256), unique=True, index=True)
    identity = Column(String(256), nullable=True)
    description = Column(String(128), nullable=True)
    can_encrypt = Column(Boolean, nullable=False)
    received_at = Column(REAL, default=time.time())


class OwnedKeys(Base):
    __tablename__ = 'owned_key'
    id = Column(Integer, primary_key=True, autoincrement=True, index=True)
    alias = Column(String(128), unique=True, index=True, nullable=True)
    private_key = Column(String(256), unique=True)
    public_key = Column(String(256), unique=True, index=True)
    description = Column(String(128))
    can_encrypt = Column(Boolean, nullable=False)
    generated_at = Column(REAL, default=time.time())


class BeamTable(Base):
    __tablename__ = "beam"
    id = Column(Integer, primary_key=True, autoincrement=True)
    blockchain_id = mapped_column(ForeignKey("blockchain.blockchain_id"), unique=True,
                                  index=True)
    beam_hash = Column(String(256), unique=True, index=True)
    owner_public_key = Column(String(256), index=True)
    target_public_key = Column(String(256), index=True)
    timestamp = Column(REAL, default=time.time(), nullable=False)

    @staticmethod
    async def save(beam, pub_key: str, target_key: str, conn_bc_id: int, comm_bc_id: int):
        from decentnet.modules.db.base import session_scope

        def add_or_update_beam(session, b_found=None):
            if beam.beam_id and b_found:
                if target_key == "NOT_KNOWN":
                    b_found.target_public_key = target_key
            else:
                # Add new BeamTable entry
                bdb = BeamTable(
                    blockchain_id=conn_bc_id,
                    beam_hash=beam.hash,
                    owner_public_key=pub_key,
                    target_public_key=target_key
                )
                session.add(bdb)
                return bdb

        async def update_blockchain_connections_async(session, bdb):
            # Retrieve the blockchain connections asynchronously
            blockchain_conn = await session.get(BlockchainTable, conn_bc_id)
            blockchain_comm = await session.get(BlockchainTable, comm_bc_id)

            blockchain_comm.beam_id = blockchain_conn.beam_id = beam.beam_id = bdb.id

        def update_blockchain_connections_sync(session, bdb):
            # Retrieve the blockchain connections synchronously
            blockchain_conn = session.get(BlockchainTable, conn_bc_id)
            blockchain_comm = session.get(BlockchainTable, comm_bc_id)

            blockchain_comm.beam_id = blockchain_conn.beam_id = beam.beam_id = bdb.id

        if USING_ASYNC_DB:
            async with session_scope() as session:
                result = await session.execute(
                    select(BeamTable).where(
                        (BeamTable.id == beam.beam_id) & (
                                (BeamTable.owner_public_key == pub_key) |
                                (BeamTable.target_public_key == pub_key)
                        )
                    )
                )
                b_found = result.scalar_one_or_none()
                bdb = add_or_update_beam(session, b_found)
                await session.flush()  # Ensure the database is updated

                await update_blockchain_connections_async(session, bdb)
        else:
            with session_scope() as session:
                result = session.execute(
                    select(BeamTable).where(
                        (BeamTable.id == beam.beam_id) & (
                                (BeamTable.owner_public_key == pub_key) |
                                (BeamTable.target_public_key == pub_key)
                        )
                    )
                )
                b_found = result.scalar_one_or_none()
                bdb = add_or_update_beam(session, b_found)
                session.flush()  # Ensure the database is updated

                update_blockchain_connections_sync(session, bdb)


class Edge(Base):
    __tablename__ = 'edge'
    id = Column(Integer, autoincrement=True, primary_key=True, index=True)
    pub_key = Column(String(256), nullable=False, comment="Edge public key", index=True)
    target = Column(String(256), nullable=True, comment="Target public key", index=True)
    capacity = Column(Integer, nullable=False)
    difficulty = Column(Integer, nullable=False,
                        default=0)
    connected = Column(Boolean, nullable=False, default=True)
    added_at = Column(REAL, default=time.time())
    # Ensure that the combination of pub_key and target is unique
    __table_args__ = (UniqueConstraint('pub_key', 'target', name='pub_target_uk'),)


class NodeInfoTable(Base):
    __tablename__ = "node_info"
    # beacon or edge id mapped
    id = Column(Integer, primary_key=True, autoincrement=True)
    edge_id = mapped_column(ForeignKey("edge.id"), unique=True,
                            index=True, nullable=True)
    ipv4 = Column(String(64), nullable=True)
    port = Column(Integer, nullable=True)
    ipv6 = Column(String(128), nullable=True)
    mac = Column(String(17),
                 nullable=True)
    pub_key = Column(String(256), nullable=False, index=True, unique=True)
    last_ping = Column(REAL, default=time.time())


def one_hour_from_now():
    return datetime.now(UTC) + timedelta(minutes=5)


class Mail(Base):
    __tablename__ = "mailbox"
    id = Column(Integer, primary_key=True, autoincrement=True)
    target = Column(String(256), nullable=True, index=True)
    block = Column(BLOB, nullable=False)
    request = Column(BLOB, nullable=False)
    expire_at = Column(DateTime, nullable=False, default=one_hour_from_now)


class AliveBeam(Base):
    __tablename__ = "alive_beam"
    id = Column(Integer, primary_key=True, autoincrement=True)
    pub_key = Column(String(256), nullable=False, index=True, unique=True)
    ready = Column(Boolean, nullable=False, default=False)
    added_at = Column(REAL, default=time.time())

    @classmethod
    async def mark_beam_ready(cls, pub_key: str, ready: bool):
        from decentnet.modules.db.base import session_scope

        if USING_ASYNC_DB:
            async with session_scope() as session:
                try:
                    # Asynchronously query the database
                    result = await session.execute(
                        select(cls).where(cls.pub_key == pub_key)
                    )
                    csd = result.scalar_one_or_none()

                    if not csd:
                        ab = AliveBeam(pub_key=pub_key, ready=ready)
                        session.add(ab)
                    else:
                        csd.ready = ready
                except (IntegrityError, sqlalchemy.exc.IntegrityError):
                    pass  # Handle or log the exception as needed
        else:
            with session_scope() as session:
                try:
                    # Synchronously query the database
                    result = session.execute(
                        select(cls).where(cls.pub_key == pub_key)
                    )
                    csd = result.scalar_one_or_none()

                    if not csd:
                        ab = AliveBeam(pub_key=pub_key, ready=ready)
                        session.add(ab)
                    else:
                        csd.ready = ready

                except (IntegrityError, sqlalchemy.exc.IntegrityError):
                    pass  # Handle or log the exception as needed



import math
import string
import unittest

from decentnet.modules.passgen.passgen import SecurePasswordGenerator


class TestPasswordGenerator(unittest.TestCase):
    def test_password_content(self):
        # Test parameters
        length = 40
        has_digits = True
        has_upper = True
        has_lower = True
        has_special = True

        # Instantiate the password generator
        generator = SecurePasswordGenerator(length, has_digits, has_upper, has_lower,
                                            has_special)

        # Generate the password
        password = generator.generate()

        # Assert conditions
        if has_digits:
            self.assertTrue(any(char in string.digits for char in password),
                            "Password lacks digits")
        if has_upper:
            self.assertTrue(any(char in string.ascii_uppercase for char in password),
                            "Password lacks uppercase letters")
        if has_lower:
            self.assertTrue(any(char in string.ascii_lowercase for char in password),
                            "Password lacks lowercase letters")
        if has_special:
            self.assertTrue(any(char in string.punctuation for char in password),
                            "Password lacks special characters")

        # Assert length
        self.assertEqual(len(password), length,
                         "Password length does not match the specified length")
        print(password)

    def test_password_entropy(self):
        # Test parameters
        length = 40
        has_digits = True
        has_upper = True
        has_lower = True
        has_special = True

        # Instantiate the password generator
        generator = SecurePasswordGenerator(length, has_digits, has_upper, has_lower,
                                            has_special)
        password = generator.generate()

        # Determine character set size based on selected parameters
        char_pool = ''
        if has_digits:
            char_pool += string.digits
        if has_upper:
            char_pool += string.ascii_uppercase
        if has_lower:
            char_pool += string.ascii_lowercase
        if has_special:
            char_pool += string.punctuation

        actual_charset = set(password)
        charset_size = len(set(char_pool))
        actual_charset_size = len(actual_charset)

        # Theoretical entropy (assuming uniform distribution)
        theoretical_entropy = length * math.log2(charset_size)

        # Actual entropy approximation based on character diversity
        actual_entropy = length * math.log2(actual_charset_size) if actual_charset_size > 1 else 0

        # Assert that theoretical entropy meets the threshold
        self.assertTrue(theoretical_entropy >= 128,
                        f"Theoretical password entropy too low: {theoretical_entropy:.2f} bits, should be >= 128 bits")

        # Assert that actual entropy meets the threshold
        self.assertTrue(actual_entropy >= 128,
                        f"Actual password entropy too low: {actual_entropy:.2f} bits, should be >= 128 bits")

        charset_size = len(set(char_pool))
        # Theoretical entropy (assuming uniform distribution)
        entropy = length * math.log2(charset_size)

        # Assert that entropy is above a strong threshold, for example 128 bits
        self.assertTrue(entropy >= 128,
                        f"Password entropy too low: {entropy:.2f} bits, should be >= 128 bits")

    def test_password_strength_heuristics(self):
        # Test parameters
        length = 40
        has_digits = True
        has_upper = True
        has_lower = True
        has_special = True

        # Instantiate the password generator
        generator = SecurePasswordGenerator(length, has_digits, has_upper, has_lower,
                                            has_special)
        password = generator.generate()
        # Check for no long runs of the same character
        current_run_char = None
        current_run_length = 0

        for ch in password:
            if ch == current_run_char:
                current_run_length += 1
            else:
                current_run_char = ch
                current_run_length = 1

            if current_run_length > 5:  # Arbitrary threshold: no more than 5 identical chars in a row
                self.fail(
                    "Password contains a long run of identical characters, potentially reducing strength")

        # Additional heuristic checks could be added here.
        # For now, we'll just ensure that we have at least 3 different character types present
        # when we've requested all four. This should be guaranteed by previous tests, but let's assert it:
        types_count = sum([
            any(c in string.digits for c in password) if has_digits else 0,
            any(c in string.ascii_uppercase for c in password) if has_upper else 0,
            any(c in string.ascii_lowercase for c in password) if has_lower else 0,
            any(c in string.punctuation for c in password) if has_special else 0
        ])

        self.assertTrue(types_count >= 3,
                        "Password does not contain at least three of the required character sets, reducing its complexity")


if __name__ == '__main__':
    unittest.main()



import unittest
from time import perf_counter

from decentnet.modules.cryptography.symmetric_aes import AESCipher


class TestAESCipher(unittest.TestCase):

    def test_encrypt_decrypt_with_different_key_sizes(self):
        for key_size in [128, 192, 256]:
            with self.subTest(key_size=key_size):
                password = "testpassword".encode()
                plaintext = b"This is a test message."
                cipher = AESCipher(password=password, key_size=key_size)

                # Encrypt the plaintext
                encrypted_text = cipher.encrypt(plaintext)

                # Decrypt the encrypted text
                decrypted_text = cipher.decrypt(encrypted_text)

                # Assert that the decrypted text is the same as the original plaintext
                self.assertEqual(decrypted_text, plaintext)

    def test_invalid_key_size(self):
        with self.assertRaises(ValueError):
            AESCipher(b"array", 64)  # An unsupported key size

    def test_performance(self):
        password = "testpassword".encode()
        key_size = 256
        plaintext = b"A" * (4 * 1024 * 1024)  # 4 MB of data
        cipher = AESCipher(password=password, key_size=key_size)

        num_runs = 30  # Number of runs to average
        encryption_times = []
        decryption_times = []

        for _ in range(num_runs):
            # Measure encryption performance
            start_time = perf_counter()
            encrypted_text = cipher.encrypt(plaintext)
            encryption_times.append((perf_counter() - start_time) * 1000)  # Convert to milliseconds

            # Measure decryption performance
            start_time = perf_counter()
            decrypted_text = cipher.decrypt(encrypted_text)
            decryption_times.append((perf_counter() - start_time) * 1000)  # Convert to milliseconds

            # Assert decryption matches original plaintext
            self.assertEqual(decrypted_text, plaintext)

        avg_encrypt_time = sum(encryption_times) / num_runs
        avg_decrypt_time = sum(decryption_times) / num_runs

        print(f"Average encryption time over {num_runs} runs: {avg_encrypt_time:.2f} ms.")
        print(f"Average decryption time over {num_runs} runs: {avg_decrypt_time:.2f} ms.")
        self.assertLess(avg_encrypt_time, 10)
        self.assertLess(avg_decrypt_time, 10)

    def test_performance_small_data(self):
        """
        Measures the performance of encryption and decryption with smaller data (128 KB).
        """
        password = b"testpassword"
        key_size = 256
        plaintext = b"A" * (2 * 1024)  # 2 KB of data
        cipher = AESCipher(password=password, key_size=key_size)

        num_runs = 10  # Number of runs to average
        encryption_times = []
        decryption_times = []

        for _ in range(num_runs):
            # Measure encryption performance
            start_time = perf_counter()
            encrypted_text = cipher.encrypt(plaintext)
            encryption_times.append((perf_counter() - start_time) * 1000)  # Convert to milliseconds

            # Measure decryption performance
            start_time = perf_counter()
            decrypted_text = cipher.decrypt(encrypted_text)
            decryption_times.append((perf_counter() - start_time) * 1000)  # Convert to milliseconds

            # Assert decryption matches original plaintext
            self.assertEqual(decrypted_text, plaintext)

        avg_encrypt_time = sum(encryption_times) / num_runs
        avg_decrypt_time = sum(decryption_times) / num_runs

        print(f"Average encryption time for 2 KB over {num_runs} runs: {avg_encrypt_time:.2f} ms.")
        print(f"Average decryption time for 2 KB over {num_runs} runs: {avg_decrypt_time:.2f} ms.")

        self.assertLess(avg_encrypt_time, 0.1)
        self.assertLess(avg_decrypt_time, 0.1)



if __name__ == '__main__':
    unittest.main()



"""Beam constants for the consensus"""
from argon2 import Type

BEAM_HASH_SIZE = 12  # B
BEAM_AES_ENCRYPTION_KEY_SIZE = 256  # Bits
DEFAULT_HANDSHAKE_AES_KEY_SIZE = 256  # Bits
DEFAULT_AES_SALT_SIZE = 4  # B
DEFAULT_AES_GCM_NONCE_SIZE = 12  # B
DEFAULT_AES_GCM_TAG_SIZE = 16  # B
DEFAULT_AES_SALT = b"\x85\x00\x01\x08\x86\x03\x04\x09"

USED_ALGORITHM = "AES"

KEY_DERIVE_TIME_COST = 1
KEY_DERIVE_MEM_COST = 8
KEY_DERIVE_PARALLELISM_COST = 1
KEY_DERIVE_ARGON_TYPE = Type.ID

AES_FRAME_SIZE = 4



import sys
from random import uniform

import numpy as np
import pyqtgraph.opengl as gl
from PySide6 import QtCore, QtWidgets

from decentnet.modules.latency_balance.icosa_balancer import IcosaBalancer

# Constants
NUM_POINTS = 35  # Number of points in the sphere
GOLDEN_ANGLE = np.pi * (3 - np.sqrt(5))  # Golden angle in radians for Fibonacci sphere
ADJUSTMENT_SPEED = 0.4  # Fraction of the adjustment applied per cycle (5%)
NEW_TARGET_SPEED = 0.1  # Speed for transitioning to new values
DISTANCE_SCALE = 0.2  # Scale factor for distance from center
FRAME_RATE = 30  # Timer interval in milliseconds
BATCH_UPDATE_INTERVAL = 5  # Number of cycles between batch updates

# Latency ranges for initial values
LATENCY_MIN = 1
LATENCY_MAX = 150

# Point size and colors
POINT_COLOR = (0.5, 0, 0.5, 0.9)  # RGBA color for points
POINT_SIZE = 8  # Size of each point

# Line color and width
LINE_COLOR = (0.4, 0.8, 0.95, 0.2)  # RGBA color for lines
LINE_WIDTH = 0.1  # Width of each line


class PointData:
    """Class to hold per-point data for animation."""

    def __init__(self, client_id, initial_in, theta, phi):
        self.client_id = client_id
        self.in_latency = initial_in
        self.target_in = initial_in  # Initial target is the starting value
        self.adjustment = 0.0  # Incremental adjustment per cycle
        self.theta = theta
        self.phi = phi
        self.point_item = None


class Latency3DApp(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("3D Spherical Latency Visualization")
        self.setGeometry(100, 100, 800, 600)

        # Generate initial client data
        self.clients_data = {
            f"client_{i}": {
                "in_latency": uniform(LATENCY_MIN, LATENCY_MAX),
                "out_latency": uniform(LATENCY_MIN, LATENCY_MAX),
                "byte_size": 1
            }
            for i in range(NUM_POINTS)
        }

        # Set up GLViewWidget for 3D rendering
        self.view = gl.GLViewWidget()
        self.view.setCameraPosition(distance=80)
        self.setCentralWidget(self.view)

        # Timer for animation updates
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.update_animation)
        self.timer.start(FRAME_RATE)

        # Create points and lines for visualization
        self.point_data_list = []
        self.line_items = []  # Store line items for easy removal
        self.create_points_and_lines()

        # Counter to control batch updates
        self.update_counter = 0

    def fibonacci_sphere(self, num_points):
        """Generate theta and phi angles for points distributed on a sphere using the Fibonacci sphere algorithm."""
        theta_phi_pairs = []

        for i in range(num_points):
            y = 1 - (i / float(num_points - 1)) * 2  # y goes from 1 to -1
            radius = np.sqrt(1 - y * y)  # Radius at y

            theta = GOLDEN_ANGLE * i  # Golden angle increment
            phi = np.arccos(y)  # Derived from y value for spherical distribution

            theta_phi_pairs.append((theta, phi))

        return theta_phi_pairs

    def create_points_and_lines(self):
        """Create points in a spherical distribution using Fibonacci sphere for even placement."""
        theta_phi_pairs = self.fibonacci_sphere(len(self.clients_data))

        for (client_id, client_data), (theta, phi) in zip(self.clients_data.items(), theta_phi_pairs):
            initial_in = client_data["in_latency"]

            # Create PointData object with fixed theta and phi from Fibonacci sphere
            point_data = PointData(client_id, initial_in, theta, phi)

            # Calculate initial position on the sphere with distance based on initial_in
            distance = initial_in * DISTANCE_SCALE
            x = distance * np.sin(phi) * np.cos(theta)
            y = distance * np.sin(phi) * np.sin(theta)
            z = distance * np.cos(phi)

            # Create a scatter plot point
            point_item = gl.GLScatterPlotItem(pos=np.array([[x, y, z]]), color=POINT_COLOR, size=POINT_SIZE)
            point_data.point_item = point_item
            self.point_data_list.append(point_data)
            self.view.addItem(point_item)

    def clear_lines(self):
        """Remove old lines from the view to prevent performance issues."""
        for line in self.line_items:
            self.view.removeItem(line)
        self.line_items.clear()

    def update_lines(self):
        """Update lines between points based on current positions."""
        self.clear_lines()  # Remove old lines before adding new ones

        # Create lines connecting each pair of points
        for i in range(len(self.point_data_list)):
            for j in range(i + 1, len(self.point_data_list)):
                start = self.point_data_list[i].point_item.pos[0]
                end = self.point_data_list[j].point_item.pos[0]
                line = gl.GLLinePlotItem(pos=np.array([start, end]), color=LINE_COLOR, width=LINE_WIDTH,
                                         antialias=True)
                self.line_items.append(line)
                self.view.addItem(line)

    def generate_new_targets(self):
        """Generate new target latencies and set adjustments using IcosaBalancer."""
        for point_data in self.point_data_list:
            # Set new target latencies randomly
            point_data.target_in = uniform(LATENCY_MIN, LATENCY_MAX)
            self.clients_data[point_data.client_id]["in_latency"] = point_data.in_latency
            self.clients_data[point_data.client_id]["out_latency"] = point_data.target_in

        # Calculate adjustments with IcosaBalancer based on new target values
        adjustments = IcosaBalancer.balance_client_latencies(self.clients_data)

        # Assign adjustments to each point's data
        for point_data in self.point_data_list:
            point_data.adjustment = adjustments[point_data.client_id]["in_latency_adjustment"]

    def update_animation(self):
        """Animate each point moving closer or farther based on in_latency values, with adjustments to target value using IcosaBalancer."""

        # Define rotation speeds for azimuth and elevation to achieve diagonal rotation
        azimuth_speed = 0.5  # Adjust for speed of azimuth rotation
        elevation_speed = 0.3  # Adjust for speed of elevation rotation

        # Retrieve and increment the camera rotation angles for diagonal movement
        current_azimuth = self.view.opts['azimuth'] + azimuth_speed
        current_elevation = self.view.opts['elevation'] + elevation_speed

        # Apply the rotation by updating camera position with new azimuth and elevation
        self.view.setCameraPosition(azimuth=current_azimuth, elevation=current_elevation)

        # Every BATCH_UPDATE_INTERVAL cycles, generate new targets and calculate adjustments
        if self.update_counter % BATCH_UPDATE_INTERVAL == 0:
            self.generate_new_targets()

        # Smoothly transition each point towards its target latency position
        for point_data in self.point_data_list:
            distance_to_target = point_data.target_in - point_data.in_latency
            if abs(distance_to_target) > 0.1:
                point_data.in_latency += distance_to_target * NEW_TARGET_SPEED
            else:
                point_data.in_latency += point_data.adjustment * ADJUSTMENT_SPEED

            # Calculate the new position based on the updated in_latency
            distance = point_data.in_latency * DISTANCE_SCALE
            x = distance * np.sin(point_data.phi) * np.cos(point_data.theta)
            y = distance * np.sin(point_data.phi) * np.sin(point_data.theta)
            z = distance * np.cos(point_data.phi)

            # Update point position with new coordinates
            point_data.point_item.setData(pos=np.array([[x, y, z]]))

        # Increment update counter and refresh lines and view
        self.update_counter += 1
        self.update_lines()
        self.view.update()  # Refresh the view


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    window = Latency3DApp()
    window.show()
    sys.exit(app.exec())



import asyncio
import os
import socket
import threading
import unittest

from decentnet.modules.tcp.socket_functions import recv_all

# Constants for length prefix generation
BLOCK_PREFIX_LENGTH_BYTES = 4
ENDIAN_TYPE = "big"


@unittest.skipUnless(os.getenv("RUN_CONN_TESTS", "False").lower() in ("1", "true", "yes"),
                     "Skipping connection tests (RUN_CONN_TESTS not set to True)")
class TestRecvAllWithSockets(unittest.TestCase):
    HOST = "127.0.0.1"
    PORT = 0

    def setUp(self):
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.bind((self.HOST, self.PORT))
        self.server_socket.listen(1)
        self.server_host, self.server_port = self.server_socket.getsockname()

        # Shared variables for assertions in threads
        self.exception_flag = None

    def tearDown(self):
        self.server_socket.close()
        if self.exception_flag:
            raise self.exception_flag

    def start_server(self, server_handler):
        """Start the server with a specific handler."""
        self.server_thread = threading.Thread(target=server_handler, daemon=True)
        self.server_thread.start()

    def test_valid_data_transfer(self):
        # Create an event to signal that the server is ready.
        server_ready = threading.Event()

        def server_handler():
            try:
                # Signal that the server is ready to accept connections.
                server_ready.set()
                conn, _ = self.server_socket.accept()
                data, data_len = asyncio.run(recv_all(conn))
                self.assertEqual(data, b"Hello, world!")
                self.assertEqual(data_len, 13)
            except Exception as e:
                self.exception_flag = e
            finally:
                conn.close()

        self.start_server(server_handler)

        # Wait until the server is ready (or use a timeout as needed)
        if not server_ready.wait(timeout=5):
            self.fail("Server did not become ready in time")

        async def client_handler():
            client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                client_socket.connect((self.server_host, self.server_port))
                message = b"Hello, world!"
                data_len = len(message)
                length_prefix = int.to_bytes(data_len, BLOCK_PREFIX_LENGTH_BYTES, ENDIAN_TYPE, signed=False)
                client_socket.sendall(length_prefix + message)
            finally:
                client_socket.close()

        asyncio.run(client_handler())

    def test_incomplete_data(self):
        def server_handler():
            try:
                conn, _ = self.server_socket.accept()
                asyncio.run(recv_all(conn))
            except ConnectionError:
                pass  # Expected exception
            except Exception as e:
                self.exception_flag = e
            finally:
                conn.close()

        self.start_server(server_handler)

        async def client_handler():
            client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            client_socket.connect((self.server_host, self.server_port))
            try:
                message = b"Hello, world!"
                data_len = len(message)
                length_prefix = int.to_bytes(data_len, BLOCK_PREFIX_LENGTH_BYTES, ENDIAN_TYPE, signed=False)
                client_socket.sendall(length_prefix + message[:5])  # Send partial data
            finally:
                client_socket.close()

        asyncio.run(client_handler())

    def test_invalid_length_prefix(self):
        def server_handler():
            try:
                conn, _ = self.server_socket.accept()
                asyncio.run(recv_all(conn))
            except ValueError:
                pass  # Expected exception
            except Exception as e:
                self.exception_flag = e
            finally:
                conn.close()

        self.start_server(server_handler)

        async def client_handler():
            client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            client_socket.connect((self.server_host, self.server_port))
            try:
                client_socket.sendall(b"abcd")  # Invalid length prefix
            finally:
                client_socket.close()

        asyncio.run(client_handler())

    def test_oversized_payload(self):
        def server_handler():
            try:
                conn, _ = self.server_socket.accept()
                asyncio.run(recv_all(conn))
            except MemoryError:
                pass  # Expected exception
            except Exception as e:
                self.exception_flag = e
            finally:
                conn.close()

        self.start_server(server_handler)

        async def client_handler():
            client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            client_socket.connect((self.server_host, self.server_port))
            try:
                oversized_message = b"A" * (10 ** 7)  # 10 MB message
                data_len = len(oversized_message)
                length_prefix = int.to_bytes(data_len, BLOCK_PREFIX_LENGTH_BYTES, ENDIAN_TYPE, signed=False)
                client_socket.sendall(length_prefix + oversized_message)
            finally:
                client_socket.close()

        asyncio.run(client_handler())

    def test_unexpected_connection_close(self):
        def server_handler():
            try:
                conn, _ = self.server_socket.accept()
                asyncio.run(recv_all(conn))
            except ConnectionError:
                pass  # Expected exception
            except Exception as e:
                self.exception_flag = e
            finally:
                conn.close()

        self.start_server(server_handler)

        async def client_handler():
            client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            client_socket.connect((self.server_host, self.server_port))
            try:
                message = b"Hello, world!"
                data_len = len(message)
                length_prefix = int.to_bytes(data_len, BLOCK_PREFIX_LENGTH_BYTES, ENDIAN_TYPE, signed=False)
                client_socket.sendall(length_prefix)  # Send only prefix, then close
            finally:
                client_socket.close()

        asyncio.run(client_handler())


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np

from decentnet.modules.latency_balance.icosa_balancer import IcosaBalancer


class TestIcosaBalancer(unittest.TestCase):
    def test_balance_client_latencies(self):
        # Generate test data with a mix of regular and extreme values
        clients_data = {
            f"client_{i}": {
                "in_latency": np.random.uniform(50, 150),
                "out_latency": np.random.uniform(50, 150),
                "byte_size": np.random.uniform(1000, 2000)
            }
            for i in range(497)
        }

        # Add extreme values for testing
        clients_data["extreme_client_in"] = {"in_latency": 300.0, "out_latency": 100.0, "byte_size": 1600.0}
        clients_data["extreme_client_out"] = {"in_latency": 80.0, "out_latency": 300.0, "byte_size": 1700.0}
        clients_data["extreme_client_byte_size"] = {"in_latency": 70.0, "out_latency": 110.0,
                                                    "byte_size": 3000.0}

        # Run the balancer
        adjustments = IcosaBalancer.balance_client_latencies(clients_data)

        # Apply adjustments
        final_in_latencies = [
            clients_data[client]["in_latency"] + adjustments[client]["in_latency_adjustment"]
            for client in clients_data
        ]
        final_out_latencies = [
            clients_data[client]["out_latency"] + adjustments[client]["out_latency_adjustment"]
            for client in clients_data
        ]
        final_byte_sizes = [
            clients_data[client]["byte_size"] + adjustments[client]["byte_size_adjustment"]
            for client in clients_data
        ]

        # Check that final values are close to the target values
        for value in final_in_latencies:
            self.assertAlmostEqual(value, IcosaBalancer.IN_LATENCY_TARGET,
                                   delta=IcosaBalancer.TOLERANCE_LATENCY * IcosaBalancer.IN_LATENCY_TARGET)

        for value in final_out_latencies:
            self.assertAlmostEqual(value, IcosaBalancer.OUT_LATENCY_TARGET,
                                   delta=IcosaBalancer.TOLERANCE_LATENCY * IcosaBalancer.OUT_LATENCY_TARGET)

        for value in final_byte_sizes:
            self.assertAlmostEqual(value, IcosaBalancer.BYTE_SIZE_TARGET,
                                   delta=IcosaBalancer.TOLERANCE_BYTE_SIZE * IcosaBalancer.BYTE_SIZE_TARGET)


if __name__ == '__main__':
    unittest.main()



import asyncio
import socket
import unittest
from multiprocessing import Pipe
from unittest.mock import AsyncMock, patch

import cbor2

from decentnet.consensus.blockchain_params import BlockchainParams
from decentnet.consensus.cmd_enum import NetworkCmd
from decentnet.consensus.r2rcomm import BLOCK_COMMAND_PREFIX
from decentnet.modules.blockchain.block import Block
from decentnet.modules.tasks_base.r2r_comm import R2RComm
from decentnet.modules.tcp.socket_functions import recv_prefix


class TestR2RComm(unittest.TestCase):
    def setUp(self):
        # Create a multiprocessing pipe
        self.pipe_recv, self.pipe_send = Pipe(duplex=False)

        # Mock the relay object
        self.mock_relay = unittest.mock.MagicMock()
        self.mock_relay.alive = True
        self.mock_relay.beam_pipe_comm = {
            "test_process_uid": [self.pipe_recv, unittest.mock.MagicMock()]
        }
        self.mock_relay.received_beam_pub_key = "test_process_uid"
        self.mock_relay.beam.target_key = "target_key"

    @patch("decentnet.modules.tasks_base.r2r_comm.R2RComm.process_disconnect_block", new_callable=AsyncMock)
    def test_process_disconnect_block_invocation(self, mock_process_disconnect_block):
        # Arrange: Set up the mocked return value
        mock_process_disconnect_block.return_value = (
            "mock_serialized_block",
            Block(1, b"adasd", BlockchainParams.low_diff_argon, b""),
        )

        # Make relay.alive change to False after the first call
        def side_effect(*args, **kwargs):
            self.mock_relay.alive = False
            return "mock_serialized_block", Block(1, b"adasd", BlockchainParams.low_diff_argon, b"")

        mock_process_disconnect_block.side_effect = side_effect

        # Prepare data to send to the pipe
        disconnect_data = {
            "cmd": NetworkCmd.DISCONNECT_EDGE.value,
            "d": "target_key",
        }
        self.pipe_send.send(BLOCK_COMMAND_PREFIX + cbor2.dumps(disconnect_data))

        # Act: Initialize R2RComm
        R2RComm(self.mock_relay)

        # Assert: Check if process_disconnect_block was invoked
        mock_process_disconnect_block.assert_called_once_with(self.mock_relay, disconnect_data)

        # Assert: Verify that relay.alive is now False
        self.assertFalse(self.mock_relay.alive)


class TestRecvPrefix(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        """Setup a pair of connected sockets for testing."""
        self.server_sock, self.client_sock = socket.socketpair()
        self.server_sock.setblocking(False)  # Set to non-blocking for async reads
        self.client_sock.setblocking(False)

    async def asyncTearDown(self):
        """Clean up sockets."""
        self.server_sock.close()
        self.client_sock.close()

    async def test_recv_prefix(self):
        """Test that recv_prefix correctly receives the expected number of bytes."""
        prefix_data = b"\x00\x10"  # Example 2-byte prefix (16 in decimal)
        self.client_sock.sendall(prefix_data)

        # Run recv_prefix and verify the received data
        received_view = await recv_prefix(self.server_sock, len(prefix_data))
        self.assertEqual(received_view.tobytes(), prefix_data)

    async def test_recv_prefix_incomplete_data(self):
        """Test that recv_prefix raises ConnectionError if the socket closes early."""
        partial_data = b"\x00"  # Only 1 byte sent instead of 2
        self.client_sock.sendall(partial_data)
        await asyncio.sleep(0.1)  # Ensure send is processed
        self.client_sock.close()  # Simulate early close

        with self.assertRaises(ConnectionError):
            await recv_prefix(self.server_sock, 2)

    async def test_recv_prefix_zero_length(self):
        """Test that recv_prefix correctly returns an empty memoryview for zero-length input."""
        received_view = await recv_prefix(self.server_sock, 0)
        self.assertEqual(len(received_view), 0)


if __name__ == "__main__":
    unittest.main()



import asyncio
import os
import unittest
from random import randint

from decentnet.consensus.blockchain_params import BlockchainParams
from decentnet.modules.blockchain.block import Block
from decentnet.modules.blockchain.blockchain import Blockchain
from decentnet.modules.cryptography.symmetric_aes import AESCipher


class TestBlockSerialization(unittest.TestCase):

    def test_block_to_from_bytes(self):
        # Create a block
        b = Blockchain("test_chain")
        asyncio.run(b.finish_blockchain_init())

        # Serialize to bytes
        block_bytes = asyncio.run(b.get_last().to_bytes())

        # Deserialize from bytes
        deserialized_block = Block.from_bytes(block_bytes)

        # Assert that the deserialized block matches the original block
        self.assertEqual(b.get_last().index, deserialized_block.index)
        self.assertEqual(b.get_last().previous_hash, deserialized_block.previous_hash)
        self.assertEqual(b.get_last().diff, deserialized_block.diff)
        self.assertEqual(b.get_last().data, deserialized_block.data)
        self.assertEqual(b.get_last().nonce, deserialized_block.nonce)
        self.assertEqual(b.get_last().timestamp, deserialized_block.timestamp)

    def test_block_encrypted_data_to_from_bytes(self):
        # Create a block
        b = Blockchain("test_chain")
        asyncio.run(b.finish_blockchain_init())

        password = "testpassword".encode()
        plaintext = b"This is a test message."
        cipher = AESCipher(password=password, key_size=256)

        # Encrypt the plaintext
        encrypted_text = cipher.encrypt(plaintext)
        bt = b.template_next_block(encrypted_text)
        bt.mine()
        self.assertTrue(asyncio.run(b.insert(bt)))
        # Serialize to bytes
        block_bytes = asyncio.run(bt.to_bytes())

        # Deserialize from bytes
        deserialized_block = Block.from_bytes(block_bytes)
        self.assertEqual(len(encrypted_text), len(deserialized_block.data),
                         "Encrypted text length does not match block data length which should be same")
        deserialized_block.data = cipher.decrypt(deserialized_block.data)

        # Assert that the deserialized block matches the original block
        self.assertEqual(b.get_last().index, deserialized_block.index)
        self.assertEqual(b.get_last().previous_hash, deserialized_block.previous_hash)
        self.assertEqual(b.get_last().diff, deserialized_block.diff)
        self.assertEqual(plaintext, deserialized_block.data)
        self.assertEqual(b.get_last().nonce, deserialized_block.nonce)
        self.assertEqual(b.get_last().timestamp, deserialized_block.timestamp)

    def test_block_to_from_bytes_dynamic(self):
        difficulty = BlockchainParams.seed_difficulty  # Assume a valid Difficulty object can be instantiated like this

        # Create a block
        test_blocks = int(os.getenv("TEST_BLOCKS", 500))
        for i in range(test_blocks):
            data = os.urandom(512 + i)
            non_empty_bytes_by4 = randint(1, 8)
            prev_hash = bytes.fromhex("00" * (16 - non_empty_bytes_by4)) + bytes.fromhex(
                'abcd' * non_empty_bytes_by4)

            block = Block(i, prev_hash, difficulty, data)

            # Serialize to bytes
            block_bytes = asyncio.run(block.to_bytes())

            # Deserialize from bytes
            deserialized_block = Block.from_bytes(block_bytes)

            # Assert that the deserialized block matches the original block
            self.assertEqual(block.index, deserialized_block.index)
            self.assertEqual(block.previous_hash, deserialized_block.previous_hash)
            self.assertEqual(block.diff, deserialized_block.diff)
            self.assertEqual(block.data, deserialized_block.data)
            self.assertEqual(block.nonce, deserialized_block.nonce)
            self.assertEqual(block.timestamp, deserialized_block.timestamp)


if __name__ == '__main__':
    unittest.main()



import asyncio
import logging
import os
from time import sleep

from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.modules.db.models import AliveBeam
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)


class Consumer:
    def __init__(self, relay, beams: dict):
        while not relay.received_beam_pub_key:
            sleep(0.5)
        self.id = relay.received_beam_pub_key

        asyncio.run(self.mark_ready(False))
        self.beams = beams

        self.client = relay.beam.client

        # while not relay.beam.connected:
        #    logger.debug(f"Consumer waiting for beam {self.id} to connect")
        #    sleep(0.5)

        asyncio.run(self.mark_ready(True))
        asyncio.run(self.start())

    async def mark_ready(self, ready: bool):
        await AliveBeam.mark_beam_ready(self.id, ready)

    async def start(self):
        logger.info(f"Consumer subscribed on channel {self.id} PID: {os.getpid()}")

        while True:
            msg = self.beams[self.id][0].recv()
            await self.process_message(msg)

    async def process_message(self, msg: bytes):
        logger.debug(
            f"Consumer {self.id} sending message to {self.client.host}:{self.client.port}")
        if msg is None:
            logger.error(f"Consumer {self.id} received None message")
            return
        await self.client.send_message(msg, ack=False)



import unittest
from time import perf_counter

from decentnet.modules.cryptography.symmetric_chacha import ChaChaCipher


class TestChaChaCipher(unittest.TestCase):

    def test_encrypt_decrypt(self):
        password = "testpassword".encode()
        plaintext = b"This is a test message."
        cipher = ChaChaCipher(password=password)

        # Encrypt the plaintext
        encrypted_text = cipher.encrypt(plaintext)

        # Decrypt the encrypted text
        decrypted_text = cipher.decrypt(encrypted_text)

        # Assert that the decrypted text is the same as the original plaintext
        self.assertEqual(decrypted_text, plaintext)

    def test_performance(self):
        password = "testpassword".encode()
        plaintext = b"A" * (4 * 1024 * 1024)  # 4 MB of data
        cipher = ChaChaCipher(password=password)

        num_runs = 30  # Number of runs to average
        encryption_times = []
        decryption_times = []

        for _ in range(num_runs):
            # Measure encryption performance
            start_time = perf_counter()
            encrypted_text = cipher.encrypt(plaintext)
            encryption_times.append((perf_counter() - start_time) * 1000)  # Convert to milliseconds

            # Measure decryption performance
            start_time = perf_counter()
            decrypted_text = cipher.decrypt(encrypted_text)
            decryption_times.append((perf_counter() - start_time) * 1000)  # Convert to milliseconds

            # Assert decryption matches original plaintext
            self.assertEqual(decrypted_text, plaintext)

        avg_encrypt_time = sum(encryption_times) / num_runs
        avg_decrypt_time = sum(decryption_times) / num_runs

        print(f"Average encryption time over {num_runs} runs: {avg_encrypt_time:.2f} ms.")
        print(f"Average decryption time over {num_runs} runs: {avg_decrypt_time:.2f} ms.")
        self.assertLess(avg_encrypt_time, 30)
        self.assertLess(avg_decrypt_time, 30)

    def test_performance_small_data(self):
        password = b"testpassword"
        plaintext = b"A" * (2 * 1024)  # 2 KB of data
        cipher = ChaChaCipher(password=password)

        num_runs = 10  # Number of runs to average
        encryption_times = []
        decryption_times = []

        for _ in range(num_runs):
            # Measure encryption performance
            start_time = perf_counter()
            encrypted_text = cipher.encrypt(plaintext)
            encryption_times.append((perf_counter() - start_time) * 1000)  # Convert to milliseconds

            # Measure decryption performance
            start_time = perf_counter()
            decrypted_text = cipher.decrypt(encrypted_text)
            decryption_times.append((perf_counter() - start_time) * 1000)  # Convert to milliseconds

            # Assert decryption matches original plaintext
            self.assertEqual(decrypted_text, plaintext)

        avg_encrypt_time = sum(encryption_times) / num_runs
        avg_decrypt_time = sum(decryption_times) / num_runs

        print(f"Average encryption time for 2 KB over {num_runs} runs: {avg_encrypt_time:.2f} ms.")
        print(f"Average decryption time for 2 KB over {num_runs} runs: {avg_decrypt_time:.2f} ms.")

        self.assertLess(avg_encrypt_time, 2)
        self.assertLess(avg_decrypt_time, 2)

    def test_compromised_data(self):
        password = "testpassword".encode()
        plaintext = b"This is a test message."
        cipher = ChaChaCipher(password=password)

        # Encrypt the plaintext
        encrypted_text = cipher.encrypt(plaintext)

        # Modify the encrypted text to compromise it
        compromised_text = bytearray(encrypted_text)
        compromised_text[-1] ^= 0x01  # Flip the last byte to corrupt the data

        # Decrypting compromised text should raise a ValueError
        with self.assertRaises(ValueError):
            cipher.decrypt(bytes(compromised_text))


if __name__ == '__main__':
    unittest.main()



import time
import unittest
from datetime import datetime


class TestTimestampPerformance(unittest.TestCase):

    def setUp(self):
        self.iterations = 10000000  # Define iterations as a class attribute

    def test_datetime_now_timestamp(self):
        block_timestamp = datetime.now().timestamp()

        start_time = time.time()
        for _ in range(self.iterations):
            (datetime.now().timestamp() - block_timestamp) * 1000
        end_time = time.time()

        duration = end_time - start_time
        print(f"datetime.now().timestamp() method took {duration:.6f} seconds")

    def test_time_time(self):
        block_timestamp = time.time()

        start_time = time.time()
        for _ in range(self.iterations):
            (time.time() - block_timestamp) * 1000
        end_time = time.time()

        duration = end_time - start_time
        print(f"time.time() method took {duration:.6f} seconds")

    def test_time_monotonic(self):
        block_timestamp = time.time()

        start_time = time.time()
        for _ in range(self.iterations):
            (time.time() - block_timestamp) * 1000
        end_time = time.time()

        duration = end_time - start_time
        print(f"time.time() method took {duration:.6f} seconds")

    def test_time_perf_counter(self):
        block_timestamp = time.perf_counter()

        start_time = time.time()
        for _ in range(self.iterations):
            (time.perf_counter() - block_timestamp) * 1000
        end_time = time.time()

        duration = end_time - start_time
        print(f"time.perf_counter() method took {duration:.6f} seconds")


if __name__ == "__main__":
    unittest.main()



import numpy as np
from matplotlib import pyplot as plt

from decentnet.modules.latency_balance.icosa_balancer import IcosaBalancer

np.random.seed(9595)
num_clients = 1000
clients_data = {
    f"client_{i}": {
        "in_latency": np.random.uniform(1, 500),
        "out_latency": np.random.uniform(1, 150),
        "byte_size": np.random.uniform(150, 60000)
    }
    for i in range(num_clients)
}

# Add extreme values for testing
clients_data["extreme_client_in"] = {"in_latency": 1.0, "out_latency": 100.0, "byte_size": 9600.0}
clients_data["extreme_client_out"] = {"in_latency": 1.0, "out_latency": 300.0, "byte_size": 9700.0}
clients_data["extreme_client_byte_size"] = {"in_latency": 1.0, "out_latency": 1.0, "byte_size": 3000.0}

# Perform balancing adjustments
adjustments = IcosaBalancer.balance_client_latencies(clients_data)

# Apply adjustments
adjusted_in_latency = [clients_data[client]["in_latency"] + adjustments[client]["in_latency_adjustment"]
                       for client in clients_data]
adjusted_out_latency = [clients_data[client]["out_latency"] + adjustments[client]["out_latency_adjustment"]
                        for client in clients_data]
adjusted_byte_size = [clients_data[client]["byte_size"] + adjustments[client]["byte_size_adjustment"]
                      for client in clients_data]

# Original values
original_in_latency = [client["in_latency"] for client in clients_data.values()]
original_out_latency = [client["out_latency"] for client in clients_data.values()]
original_byte_size = [client["byte_size"] for client in clients_data.values()]

# Plot histograms for before and after adjustment
fig, axs = plt.subplots(3, 2, figsize=(12, 12))
fig.suptitle("Histograms of Latency and Byte Size Values Before and After Adjustment")

# Plot in_latency
axs[0, 0].hist(original_in_latency, bins=20, alpha=0.7, color='blue')
axs[0, 0].set_title("In Latency (Before Adjustment)")
axs[0, 1].hist(adjusted_in_latency, bins=20, alpha=0.7, color='green')
axs[0, 1].set_title("In Latency (After Adjustment)")

# Plot out_latency
axs[1, 0].hist(original_out_latency, bins=20, alpha=0.7, color='blue')
axs[1, 0].set_title("Out Latency (Before Adjustment)")
axs[1, 1].hist(adjusted_out_latency, bins=20, alpha=0.7, color='green')
axs[1, 1].set_title("Out Latency (After Adjustment)")

# Plot byte_size
axs[2, 0].hist(original_byte_size, bins=20, alpha=0.7, color='blue')
axs[2, 0].set_title("Byte Size (Before Adjustment)")
axs[2, 1].hist(adjusted_byte_size, bins=20, alpha=0.7, color='green')
axs[2, 1].set_title("Byte Size (After Adjustment)")

plt.tight_layout(rect=[0, 0, 1, 0.96])
plt.show()



import logging
import os

from nacl.exceptions import BadSignatureError
from nacl.public import Box, PrivateKey, PublicKey
from nacl.signing import SigningKey, VerifyKey

from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.modules.convert.byte_to_base64_utils import (base64_to_original,
                                                            bytes_to_base64)
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)


class AsymCrypt:

    @staticmethod
    def generate_key_pair_encryption() -> [bytes, bytes]:
        """Generate a key pair for encryption using PyNaCl (Curve25519)."""
        private_key = PrivateKey.generate()
        public_key = private_key.public_key
        return private_key.encode(), public_key.encode()

    @staticmethod
    def generate_key_pair_signing() -> [SigningKey, VerifyKey]:
        """Generate a key pair for Ed25519 (replaces ECDSA)."""
        # Generate a private key (SigningKey)
        private_key = SigningKey.generate()

        # Derive the public key from the private key
        public_key = private_key.verify_key

        return private_key, public_key

    @staticmethod
    def check_encryption_key(public_key: bytes):
        try:
            # Ensure public key is a valid NaCl public key
            PublicKey(public_key)
            return True
        except Exception:
            return False

    @staticmethod
    def encrypt_message(public_key: bytes, data: bytes) -> bytes:
        """
        Encrypts data using PyNaCl Box (public-key encryption).
        :param public_key: The recipient's public key (as bytes).
        :param data: The plaintext data to encrypt (as bytes).
        :return: Encrypted data with nonce included.
        """
        recipient_public_key = PublicKey(public_key)

        # Sender uses their own fixed private key (assumed to be managed externally)
        sender_private_key = PrivateKey.generate()  # This can be passed into the function instead
        box = Box(sender_private_key, recipient_public_key)

        # Encrypt the message, including a random nonce
        encrypted = box.encrypt(data, nonce=os.urandom(Box.NONCE_SIZE))

        # Prepend the sender's public key (required for decryption)
        return sender_private_key.public_key.encode() + encrypted

    @staticmethod
    def decrypt_message(private_key: bytes, encrypted_data: bytes) -> bytes:
        """
        Decrypts data using PyNaCl Box (public-key encryption).
        :param private_key: The recipient's private key (as bytes).
        :param encrypted_data: The encrypted data with nonce and sender's public key prepended.
        :return: Decrypted plaintext data.
        """
        # Extract sender's public key (first 32 bytes)
        sender_public_key = PublicKey(encrypted_data[:32])
        encrypted_message = encrypted_data[32:]

        # Use recipient's private key
        recipient_private_key = PrivateKey(private_key)

        # Create a Box with recipient's private key and sender's public key
        box = Box(recipient_private_key, sender_public_key)

        # Decrypt the message
        decrypted_data = box.decrypt(encrypted_message)
        return decrypted_data

    @staticmethod
    def sign_message(private_key: SigningKey, data: bytes) -> bytes:
        """Signs data using Ed25519 (replaces ECDSA)."""
        signature = private_key.sign(data).signature
        return signature

    @staticmethod
    def verify_signature(*, public_key: VerifyKey, signature: bytes, data: bytes) -> bool:
        """Verifies the signature of data using Ed25519."""
        try:
            public_key.verify(data, signature)
            return True
        except BadSignatureError:
            return False

    @staticmethod
    def key_pair_from_private_key_base64(private_key: str, can_encrypt: bool) -> tuple[
                                                                                     SigningKey, VerifyKey] | \
                                                                                 tuple[bytes, bytes]:
        """
        Returns an Object key pair from base64 str.
        :param private_key:
        :param can_encrypt:
        :return: Private Key Object, Public Key object.
        """
        if not can_encrypt:
            private_key = SigningKey(base64_to_original(private_key))
            public_key = private_key.verify_key
            return private_key, public_key
        else:
            pk_bytes = base64_to_original(private_key)
            private_key = PrivateKey(pk_bytes)
            public_key = private_key.public_key
            return private_key.encode(), public_key.encode()

    @staticmethod
    def public_key_from_base64(public_key: str, can_encrypt: bool) -> VerifyKey | bytes:
        """
        Returns Object key from base64 str.
        :param public_key:
        :param can_encrypt:
        :return: Private Key Object, Public Key object.
        """
        if not can_encrypt:
            return VerifyKey(base64_to_original(public_key))
        else:
            pk_bytes = base64_to_original(public_key)
            return PublicKey(pk_bytes).encode()

    @staticmethod
    def encryption_key_to_base64(key: bytes) -> str:
        """
        Converts bytes of public or private key to base64.
        :param key:
        :return:
        """
        return bytes_to_base64(key)

    @staticmethod
    def verifying_key_to_string(key: VerifyKey) -> str:
        """
        Converts a verification key into string base64.
        :param key:
        :return:
        """
        return bytes_to_base64(key.encode())

    @staticmethod
    def verifying_key_from_string(key: str) -> VerifyKey:
        return VerifyKey(base64_to_original(key))
